
import numpy as np
import pytest

import astropy.units as u
from astropy.coordinates import BaseCoordinateFrame, SkyCoord
from astropy.coordinates.earth import EarthLocation
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time
from astropy.wcs import WCS

import sunpy.map
from sunpy.coordinates.frames import (
    Heliocentric,
    HeliographicCarrington,
    HeliographicStonyhurst,
    Helioprojective,
    SunPyBaseCoordinateFrame,
)
from sunpy.coordinates.wcs_utils import (
    _set_wcs_aux_obs_coord,
    solar_frame_to_wcs_mapping,
    solar_wcs_frame_mapping,
)


@pytest.mark.parametrize(('ctype', 'frame'), [(['HPLN', 'HPLT'], Helioprojective),
                                              (['HPLT', 'HPLN'], Helioprojective),
                                              (['HGLN', 'HGLT'], HeliographicStonyhurst),
                                              (['CRLN', 'CRLT'], HeliographicCarrington),
                                              (['SOLX', 'SOLY'], Heliocentric)
                                              ])
def test_wcs_frame_mapping(ctype, frame):
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ctype
    result = solar_wcs_frame_mapping(wcs)
    assert isinstance(result, frame)


def test_wcs_frame_mapping_none():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ['spam', 'eggs']

    result = solar_wcs_frame_mapping(wcs)

    assert result is None


def test_wcs_frame_mapping_dateavg():
    wcs = WCS(naxis=2)
    wcs.wcs.dateavg = "2020-01-01T00:00:00"
    wcs.wcs.ctype = ["HPLN", "HPLT"]

    result = solar_wcs_frame_mapping(wcs)

    assert result.obstime == Time("2020-01-01T00:00:00")

    wcs = WCS(naxis=2)
    wcs.wcs.dateavg = "2020-01-01T00:00:00"
    wcs.wcs.dateobs = "2020-01-01T01:00:00"
    wcs.wcs.ctype = ["HPLN", "HPLT"]

    result = solar_wcs_frame_mapping(wcs)

    assert result.obstime == Time("2020-01-01T00:00:00")


def test_wcs_frame_mapping_observer_hgc_self():
    # Test whether a WCS with HGC coordinates for the observer location uses observer="self"
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ['SOLX', 'SOLY']
    wcs.wcs.dateobs = '2001-01-01'
    wcs.wcs.aux.crln_obs = 10
    wcs.wcs.aux.hglt_obs = 20
    wcs.wcs.aux.dsun_obs = 1.5e11

    # This frame will have the observer location in HGS
    result = solar_wcs_frame_mapping(wcs)

    # We perform the desired transformation using observer="self"
    hgc_obs = HeliographicCarrington(wcs.wcs.aux.crln_obs * u.deg,
                                     wcs.wcs.aux.hglt_obs * u.deg,
                                     wcs.wcs.aux.dsun_obs * u.m,
                                     obstime=wcs.wcs.dateobs, observer="self")
    hgs_obs = hgc_obs.transform_to(HeliographicStonyhurst(obstime=hgc_obs.obstime))

    assert_quantity_allclose(result.observer.lon, hgs_obs.lon)
    assert_quantity_allclose(result.observer.lat, hgs_obs.lat)
    assert_quantity_allclose(result.observer.radius, hgs_obs.radius)


def test_wcs_aux():
    """
    Make sure auxiliary information round trips properly from coordinate frames
    to WCS and back.
    """
    data = np.ones([6, 6], dtype=np.float64)
    header = {'CRVAL1': 0,
              'CRVAL2': 0,
              'CRPIX1': 5,
              'CRPIX2': 5,
              'CDELT1': 10,
              'CDELT2': 10,
              'CUNIT1': 'arcsec',
              'CUNIT2': 'arcsec',
              'PC1_1': 0,
              'PC1_2': -1,
              'PC2_1': 1,
              'PC2_2': 0,
              'NAXIS1': 6,
              'NAXIS2': 6,
              'CTYPE1': 'HPLN-TAN',
              'CTYPE2': 'HPLT-TAN',
              'date-obs': '1970-01-01T00:00:00',
              'obsrvtry': 'Foo',
              'detector': 'bar',
              'wavelnth': 10,
              'waveunit': 'm',
              'hglt_obs': 0,
              'hgln_obs': 0,
              'dsun_obs': 10,
              'rsun_ref': 690000000}
    generic_map = sunpy.map.Map((data, header))

    wcs = generic_map.wcs
    assert wcs.wcs.aux.hglt_obs == 0
    assert wcs.wcs.aux.hgln_obs == 0
    assert wcs.wcs.aux.dsun_obs == 10
    assert wcs.wcs.aux.rsun_ref == header['rsun_ref']

    result = solar_wcs_frame_mapping(wcs)

    assert isinstance(result, Helioprojective)
    assert result.observer.lat.value == 0
    assert result.observer.lon.value == 0
    assert result.observer.radius.value == 10
    assert result.observer.rsun.to_value(u.m) == header['rsun_ref']
    assert result.rsun.to_value(u.m) == header['rsun_ref']


def test_hpc_frame_to_wcs():
    frame = Helioprojective(observer="earth", obstime='2013-10-28', rsun=690*u.Mm)
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'HPLN-TAN'
    assert result_wcs.wcs.cunit[0] == 'arcsec'
    assert result_wcs.wcs.dateobs == '2013-10-28T00:00:00.000'
    assert result_wcs.wcs.aux.rsun_ref == frame.rsun.to_value(u.m)

    new_frame = solar_wcs_frame_mapping(result_wcs)
    assert new_frame.rsun == frame.rsun

    # Test a frame with no obstime and no observer
    frame = Helioprojective()
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'HPLN-TAN'
    assert result_wcs.wcs.cunit[0] == 'arcsec'
    assert result_wcs.wcs.dateobs == ''

    new_frame = solar_wcs_frame_mapping(result_wcs)
    assert new_frame.observer is None
    assert new_frame.rsun == frame.rsun


def test_hgs_frame_to_wcs():
    frame = HeliographicStonyhurst(obstime='2013-10-28', rsun=690*u.Mm)
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'HGLN-TAN'
    assert result_wcs.wcs.cunit[0] == 'deg'
    assert result_wcs.wcs.dateobs == '2013-10-28T00:00:00.000'
    assert result_wcs.wcs.aux.rsun_ref == frame.rsun.to_value(u.m)

    new_frame = solar_wcs_frame_mapping(result_wcs)
    assert new_frame.rsun == frame.rsun

    # Test a frame with no obstime
    frame = HeliographicStonyhurst()
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'HGLN-TAN'
    assert result_wcs.wcs.cunit[0] == 'deg'
    assert result_wcs.wcs.dateobs == ''


def test_hgc_frame_to_wcs():
    frame = HeliographicCarrington(obstime='2013-10-28', rsun=690*u.Mm)
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'CRLN-TAN'
    assert result_wcs.wcs.cunit[0] == 'deg'
    assert result_wcs.wcs.dateobs == '2013-10-28T00:00:00.000'
    assert result_wcs.wcs.aux.rsun_ref == frame.rsun.to_value(u.m)

    new_frame = solar_wcs_frame_mapping(result_wcs)
    assert new_frame.rsun == frame.rsun

    # Test a frame with no obstime
    frame = HeliographicCarrington()
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'CRLN-TAN'
    assert result_wcs.wcs.cunit[0] == 'deg'
    assert result_wcs.wcs.dateobs == ''


def test_hcc_frame_to_wcs():
    frame = Heliocentric(observer="earth", obstime='2013-10-28')
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'SOLX'
    assert result_wcs.wcs.dateobs == '2013-10-28T00:00:00.000'

    new_frame = solar_wcs_frame_mapping(result_wcs)
    assert isinstance(new_frame.observer, HeliographicStonyhurst)

    # Test a frame with no obstime and no observer
    frame = Heliocentric()
    result_wcs = solar_frame_to_wcs_mapping(frame)

    assert isinstance(result_wcs, WCS)

    assert result_wcs.wcs.ctype[0] == 'SOLX'
    assert result_wcs.wcs.dateobs == ''

    new_frame = solar_wcs_frame_mapping(result_wcs)
    assert new_frame.observer is None


def test_non_sunpy_frame_to_wcs():
    # For a non-SunPy frame, our mapping should return None
    frame = BaseCoordinateFrame()
    assert solar_frame_to_wcs_mapping(frame) is None


def test_attribute_errors():
    # Check that errors are raised if we try to convert a WCS with .rsun
    # or .heliographic_observer attributes
    wcs = WCS(naxis=2)
    wcs.rsun = None
    with pytest.raises(ValueError, match='The .rsun attribute'):
        solar_wcs_frame_mapping(wcs)

    wcs = WCS(naxis=2)
    wcs.heliographic_observer = None
    with pytest.raises(ValueError, match='The .heliographic_observer attribute'):
        solar_wcs_frame_mapping(wcs)

    wcs = WCS(naxis=2)
    wcs.heliographic_observer = None
    wcs.rsun = None
    with pytest.raises(ValueError, match='The .rsun and .heliographic_observer attribute'):
        solar_wcs_frame_mapping(wcs)


def test_set_wcs_aux():
    wcs = WCS(naxis=2)
    observer = Helioprojective(observer="earth", obstime='2013-10-28').observer
    _set_wcs_aux_obs_coord(wcs, observer)
    assert wcs.wcs.aux.hgln_obs == 0
    assert u.allclose(wcs.wcs.aux.hglt_obs, 4.7711570596394015)
    assert u.allclose(wcs.wcs.aux.dsun_obs, 148644585949.4918)
    assert wcs.wcs.aux.crln_obs is None

    wcs = WCS(naxis=2)
    observer = observer.transform_to(HeliographicCarrington(observer=observer))
    _set_wcs_aux_obs_coord(wcs, observer)
    assert wcs.wcs.aux.hgln_obs is None
    assert u.allclose(wcs.wcs.aux.hglt_obs, 4.7711570596394015)
    assert u.allclose(wcs.wcs.aux.dsun_obs, 148644585949.4918)
    assert u.allclose(wcs.wcs.aux.crln_obs, 326.05139910339886)

    observer = observer.transform_to(Heliocentric(observer=observer))
    with pytest.raises(ValueError, match='obs_coord must be in a Stonyhurst or Carrington frame'):
        _set_wcs_aux_obs_coord(wcs, observer)


def test_self_observer():
    frame = HeliographicCarrington(0*u.deg, 0*u.deg, 1*u.au,
                                   observer="self", obstime='2013-10-28')
    wcs = solar_frame_to_wcs_mapping(frame)
    assert wcs.wcs.aux.hgln_obs is None
    assert u.allclose(wcs.wcs.aux.hglt_obs, frame.lon.to_value(u.deg))
    assert u.allclose(wcs.wcs.aux.crln_obs, frame.lon.to_value(u.deg))
    assert u.allclose(wcs.wcs.aux.dsun_obs, frame.radius.to_value(u.m))


@pytest.fixture
def dkist_location():
    return EarthLocation(*(-5466045.25695494, -2404388.73741278, 2242133.88769004) * u.m)


@pytest.mark.remote_data
def test_obsgeo_frame_mapping_cartesian(dkist_location, caplog):

    obstime = Time("2021-05-21T03:00:00")
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ['HPLT', 'HPLN']
    wcs.wcs.obsgeo = list(dkist_location.to_value(u.m).tolist()) + [0, 0, 0]
    wcs.wcs.dateobs = obstime.isot

    frame = solar_wcs_frame_mapping(wcs)

    assert frame.observer is not None

    assert frame.observer == SkyCoord(dkist_location.get_itrs(obstime)).transform_to('heliographic_stonyhurst').frame

    assert not caplog.records


@pytest.mark.remote_data
def test_frame_mapping_obsgeo_spherical(dkist_location, caplog):

    obstime = Time("2021-05-21T03:00:00")
    location = dkist_location.get_itrs(obstime)
    loc_sph = location.spherical
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ['HPLT', 'HPLN']
    wcs.wcs.obsgeo = [0, 0, 0] + [loc_sph.lon.to_value(u.deg), loc_sph.lat.to_value(u.deg), loc_sph.distance.to_value(u.m)]
    wcs.wcs.dateobs = obstime.isot

    frame = solar_wcs_frame_mapping(wcs)

    assert frame.observer is not None

    assert frame.observer == SkyCoord(location).transform_to('heliographic_stonyhurst').frame

    assert not caplog.records


def test_observer_hgln_crln_priority():
    """
    When extracting the observer information from a FITS header, ensure
    Stonyhurst (HG) coordinates are preferred over Carrington (CR) if present.
    Note that `Map` creates a custom FITS header with a sanitized observer
    location, so it is separately tested in the map module.
    """
    header = {'CRVAL1': 0,
              'CRVAL2': 0,
              'CRPIX1': 5,
              'CRPIX2': 5,
              'CDELT1': 10,
              'CDELT2': 10,
              'CUNIT1': 'arcsec',
              'CUNIT2': 'arcsec',
              'PC1_1': 0,
              'PC1_2': -1,
              'PC2_1': 1,
              'PC2_2': 0,
              'NAXIS1': 6,
              'NAXIS2': 6,
              'CTYPE1': 'HPLN-TAN',
              'CTYPE2': 'HPLT-TAN',
              'date-obs': '1970-01-01T00:00:00',
              'mjd-obs': 40587,
              'hglt_obs': 0,
              'hgln_obs': 0,
              'crlt_obs': 2,
              'crln_obs': 2,
              'dsun_obs': 10,
              'rsun_ref': 690000000}
    wcs = WCS(header)
    c = wcs.pixel_to_world(0, 0)
    assert c.observer.lon == 0 * u.deg
    # Note: don't test whether crlt or hglt is used---according to
    # _set_wcs_aux_obs_coord, those are expected to always be the same and so
    # the same one is always used


def test_sunpybaseframe_external():
    class MyFrame(SunPyBaseCoordinateFrame):
        pass

    out = solar_frame_to_wcs_mapping(MyFrame())
    assert out is None



from asv_runner.benchmarks.mark import SkipNotImplemented

import astropy.units as u
from astropy.coordinates import HCRS, ITRS, HeliocentricMeanEcliptic, SphericalRepresentation

import sunpy.coordinates.frames as f


class TransformationHeliographic:
    frame_names = ['HCRS', 'HGS', 'HGC', 'HCC', 'HPC', 'HPR', 'HCI']

    params = (frame_names, frame_names)
    param_names = ['src', 'dest']

    def setup_cache(self):
        obstime = '2023-01-01'
        vect = SphericalRepresentation(0*u.deg, 0*u.deg, 1*u.AU)
        observer = f.HeliographicStonyhurst(vect, obstime=obstime)
        frames = {
            'HCRS': HCRS(vect, obstime=obstime),
            'HGS': f.HeliographicStonyhurst(vect, obstime=obstime),
            'HGC': f.HeliographicCarrington(vect, obstime=obstime, observer=observer),
            'HCC': f.Heliocentric(vect, obstime=obstime, observer=observer),
            'HPC': f.Helioprojective(vect, obstime=obstime, observer=observer),
            'HPR': f.HelioprojectiveRadial(vect, obstime=obstime, observer=observer),
            'HCI': f.HeliocentricInertial(vect, obstime=obstime),
        }
        return frames

    def setup(self, frames, src, dest):
        if src == dest:
            raise SkipNotImplemented

    def time_transform(self, frames, src, dest):
        frames[src].transform_to(frames[dest])


class TransformationEcliptic:
    frame_names = ['HAE', 'HEE', 'GSE', 'GEI']

    params = (frame_names, frame_names)
    param_names = ['src', 'dest']

    def setup_cache(self):
        obstime = '2023-01-01'
        vect = SphericalRepresentation(0*u.deg, 0*u.deg, 1*u.AU)
        frames = {
            'HAE': HeliocentricMeanEcliptic(vect, obstime=obstime, equinox='J2000'),
            'HEE': f.HeliocentricEarthEcliptic(vect, obstime=obstime),
            'GSE': f.GeocentricSolarEcliptic(vect, obstime=obstime),
            'GEI': f.GeocentricEarthEquatorial(vect, obstime=obstime, equinox='J2000'),
        }
        return frames

    def setup(self, frames, src, dest):
        if src == dest:
            raise SkipNotImplemented

    def time_transform(self, frames, src, dest):
        frames[src].transform_to(frames[dest])


class TransformationMagnetic:
    frame_names = ['GEO', 'MAG', 'SM', 'GSM']

    params = (frame_names, frame_names)
    param_names = ['src', 'dest']

    def setup_cache(self):
        obstime = '2023-01-01'
        vect = SphericalRepresentation(0*u.deg, 0*u.deg, 1*u.AU)
        frames = {
            'GEO': ITRS(vect, obstime=obstime),
            'MAG': f.Geomagnetic(vect, obstime=obstime),
            'SM': f.SolarMagnetic(vect, obstime=obstime),
            'GSM': f.GeocentricSolarMagnetospheric(vect, obstime=obstime),
        }
        return frames

    def setup(self, frames, src, dest):
        if src == dest:
            raise SkipNotImplemented

    def time_transform(self, frames, src, dest):
        frames[src].transform_to(frames[dest])



import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header_obj, get_header_from_file, get_test_filepath
from sunpy.map.sources.gong import GONGMagnetogramMap
from sunpy.time import parse_time
from .helpers import _test_private_date_setters


@pytest.fixture(params=["full-date-obs", "with-time-obs"])
def gong_magnetogram(request):
    header = get_header_from_file(get_test_filepath('gong_magnetogram.header'))
    if request.param == "full-date-obs":
        time_obs =  header.pop("TIME-OBS")
        date_obs = (parse_time(f"{header['DATE-OBS']} {time_obs}", scale='tai') + 19 *u.s).utc
        header["DATE-OBS"] = date_obs
    return get_dummy_map_from_header_obj(header)


def test_fitstoGONGSynoptic(gong_magnetogram):
    """Tests the creation of GONGMagnetogramMap using FITS."""
    assert isinstance(gong_magnetogram, GONGMagnetogramMap)


def test_is_datasource_for(gong_magnetogram):
    """Test the is_datasource_for method of GONGMagnetogramMap."""
    assert gong_magnetogram.is_datasource_for(gong_magnetogram.data, gong_magnetogram.meta)


def test_observatory(gong_magnetogram):
    """Tests the observatory property of the GONGMagnetogramMap object."""
    assert gong_magnetogram.observatory == "NSO-GONG"


def test_measurement(gong_magnetogram):
    """Tests the measurement property of the GONGMagnetogramMap object."""
    assert gong_magnetogram.measurement == 676.8 * u.nm


def test_reference_date(gong_magnetogram):
    # 2025-01-09T01:51:57
    assert gong_magnetogram.reference_date.isot == "2025-01-09T01:34:58.008"


def test_date(gong_magnetogram):
    # The time in the GONG header is in a ISO format but in GPS scale
    # GPS time (date-obs) 2025-01-09T01:51:57 should be UTC 2025-01-09T01:34:58.008 (-18 seconds)
    assert gong_magnetogram.date.isot == "2025-01-09T01:34:58.008"


def test_nickname(gong_magnetogram):
    assert gong_magnetogram.nickname == "NSO-GONG, Learmonth, (AUS)"


def test_private_date_setters(gong_magnetogram):
    _test_private_date_setters(gong_magnetogram)


def test_unit(gong_magnetogram):
    assert gong_magnetogram.unit == u.G
    assert gong_magnetogram.unit == u.Unit("Mx/cm^2")
    assert gong_magnetogram.unit.to_string() == 'G'


def test_spatial_units(gong_magnetogram):
    assert gong_magnetogram.spatial_units[0] == u.deg
    assert gong_magnetogram.spatial_units[1] == u.deg


def test_rsun_obs(gong_magnetogram):
    assert u.allclose(gong_magnetogram.rsun_obs, 975.86818683 * u.arcsec)


def test_wcs(gong_magnetogram):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    gong_magnetogram.pixel_to_world(0*u.pix, 0*u.pix)



"""
=====================================
Reprojecting Using a Spherical Screen
=====================================

This example demonstrates how you can reproject an image as if it lies on the
inside of a spherical screen and the observer is not at the center of the
sphere. This functionality is primarily for visualization purposes, since
features in the image are unlikely to actually lie on this spherical screen.

You will need `reproject <https://reproject.readthedocs.io/en/stable/>`__ v0.6
or higher installed.
"""
# sphinx_gallery_thumbnail_number = 4

import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.coordinates import SphericalScreen
from sunpy.data.sample import AIA_171_IMAGE

######################################################################
# We will use one of the AIA images from the sample data. We fix the
# range of values for the Map's normalizer.

aia_map = sunpy.map.Map(AIA_171_IMAGE)
aia_map.plot_settings['norm'].vmin = 0
aia_map.plot_settings['norm'].vmax = 10000

fig = plt.figure()
ax = fig.add_subplot(projection=aia_map)
aia_map.plot(axes=ax)
plt.show()

######################################################################
# Let's define a new observer that is well separated from Earth.

new_observer = SkyCoord(70*u.deg, 20*u.deg, 1*u.AU, obstime=aia_map.date,
                        frame='heliographic_stonyhurst')

######################################################################
# Create a WCS header for this new observer using helioprojective
# coordinates.

out_shape = aia_map.data.shape

out_ref_coord = SkyCoord(0*u.arcsec, 0*u.arcsec, obstime=new_observer.obstime,
                         frame='helioprojective', observer=new_observer,
                         rsun=aia_map.coordinate_frame.rsun)
out_header = sunpy.map.make_fitswcs_header(
    out_shape,
    out_ref_coord,
    scale=u.Quantity(aia_map.scale),
    instrument=aia_map.instrument,
    wavelength=aia_map.wavelength
)

######################################################################
# If you reproject the AIA Map to the perspective of the new observer,
# the default assumption is that the image lies on the surface of the
# Sun. However, the parts of the image beyond the solar disk cannot
# be mapped to the surface of the Sun, and thus do not show up in the
# output.

outmap_default = aia_map.reproject_to(out_header)

fig = plt.figure()
ax = fig.add_subplot(projection=outmap_default)
outmap_default.plot(axes=ax)
plt.show()

######################################################################
# You can use the different assumption that the image lies on the
# surface of a spherical screen centered at AIA, with a radius equal
# to the Sun-AIA distance. The curvature of the spherical screen is
# not obvious in this plot due to the relatively small field of view
# of AIA (compared to, say, a coronagraph).

with SphericalScreen(aia_map.observer_coordinate):
    outmap_screen_all = aia_map.reproject_to(out_header)

fig = plt.figure()
ax = fig.add_subplot(projection=outmap_screen_all)
outmap_screen_all.plot(axes=ax)
plt.show()

######################################################################
# Finally, you can specify that the spherical-screen assumption should
# be used for only off-disk parts of the image, and continue to map
# on-disk parts of the image to the surface of the Sun.

with SphericalScreen(aia_map.observer_coordinate, only_off_disk=True):
    outmap_screen_off_disk = aia_map.reproject_to(out_header)

fig = plt.figure()
ax = fig.add_subplot(projection=outmap_screen_off_disk)
outmap_screen_off_disk.plot(axes=ax)
plt.show()



"""
Test cases for ASO-S HXIMap subclass.
"""
import pytest

import astropy.units as u
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.asos import HXIMap
from sunpy.util.exceptions import SunpyMetadataWarning
from .helpers import _test_private_date_setters


@pytest.fixture(scope="module", params=[0,1])
def hxi_map(request):
    return get_dummy_map_from_header(get_test_filepath(f"hxi_imgcube_01e02t_20230501_130758_HXI_CLEAN_{request.param}.header"))


def test_HXIMap(hxi_map):
    """Tests the creation of HXIMap"""
    assert isinstance(hxi_map, HXIMap)


def test_is_datasource_for(hxi_map):
    """Tests the is_datasource_for method of HXIMap."""
    assert hxi_map.is_datasource_for(hxi_map.data, hxi_map.meta)


def test_observatory(hxi_map):
    """Tests the observatory property of the HXIMap object."""
    assert hxi_map.observatory == "ASO-S"


def test_reference_date(hxi_map):
    assert hxi_map.reference_date.isot in ["2023-05-01T13:07:58.713", "2023-05-01T13:08:16.962"]


def test_date(hxi_map):
    assert hxi_map.date.isot in ["2023-05-01T13:07:58.713", "2023-05-01T13:08:16.962"]


def test_unit(hxi_map):
    assert hxi_map.unit == u.photon / (u.s * u.arcsec**2 * u.cm**2)


def test_broken_units(hxi_map):
    # Check that no unit or bunit leads to no unit
    hxi_map.meta["unit"] = None
    assert hxi_map.unit is None

    # Check that other units are parsed
    hxi_map.meta["unit"] = "second"
    assert hxi_map.unit == u.s

    # Check that it falls back to fits standard key
    hxi_map.meta["bunit"] = "second"
    assert hxi_map.unit == u.s

    # Check that non-fits units still raise
    hxi_map.meta["unit"] = "spam"
    with pytest.warns(SunpyMetadataWarning):
        hxi_map.unit


def test_private_date_setters(hxi_map):
    _test_private_date_setters(hxi_map)


def test_measurement(hxi_map):
    """Tests the measurement property of the HXIMap object."""
    assert u.allclose(hxi_map.measurement, [20, 30] * u.keV)


def test_norm_clip(hxi_map):
    # Tests that the default normalizer has clipping disabled
    assert not hxi_map.plot_settings['norm'].clip


def test_new_instance_preserves_plot_settings(hxi_map):
    # Tests that the _new_instance method preserves the plot_settings
    # of the old instance. This is done on the HXI source as the HXIMap
    # constructor explicitly sets the cmap and norm and we want to test
    # that _new_instance persists the old custom plot_settings
    hxi_map.plot_settings['norm'] = ImageNormalize(vmin=0.1, vmax=42)
    hxi_map.plot_settings['cmap'] = 'inferno'
    new_hxi_map = hxi_map._new_instance(hxi_map.data,
                                        hxi_map.meta,
                                        plot_settings=hxi_map.plot_settings)
    assert new_hxi_map.plot_settings['norm'].vmin == hxi_map.plot_settings['norm'].vmin
    assert new_hxi_map.plot_settings['norm'].vmax == hxi_map.plot_settings['norm'].vmax
    assert new_hxi_map.plot_settings['cmap'] == hxi_map.plot_settings['cmap']
    # If no plot settings are explicitly passed, the plot_settings should fall back to those
    # in the constructor
    new_hxi_map = hxi_map._new_instance(hxi_map.data, hxi_map.meta)
    assert new_hxi_map.plot_settings['norm'].vmin is None
    assert new_hxi_map.plot_settings['norm'].vmax is None
    assert new_hxi_map.plot_settings['cmap'] == new_hxi_map._get_cmap_name()


def test_wcs(hxi_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    hxi_map.pixel_to_world(0*u.pix, 0*u.pix)



"""
This module provides functions that draw on Astropy's WCSAxes.
"""
import numpy as np
from matplotlib import patches
from matplotlib.path import Path

import astropy.units as u
from astropy.constants import R_sun
from astropy.coordinates import SkyCoord
from astropy.visualization.wcsaxes.wcsapi import wcsapi_to_celestial_frame
from astropy.wcs.wcsapi import BaseHighLevelWCS

from sunpy.coordinates import HeliographicCarrington, HeliographicStonyhurst
from sunpy.coordinates.frames import HeliocentricInertial, Helioprojective
from sunpy.coordinates.sun import _angular_radius
from sunpy.coordinates.utils import get_limb_coordinates
from sunpy.util import grid_perimeter
from sunpy.visualization import wcsaxes_compat

__all__ = ["limb", "equator", "prime_meridian", "extent"]


@u.quantity_input
def limb(axes, observer, *, rsun: u.m = R_sun, resolution=1000, **kwargs):
    """
    Draws the solar limb as seen by the specified observer.

    The limb is a circle for only the simplest plots. If the specified
    observer of the limb is different from the observer of the coordinate frame
    of the plot axes, not only may the limb not be a true circle, a portion of
    the limb may be hidden from the observer. In that case, the circle is
    divided into visible and hidden segments, represented by solid and dotted
    lines, respectively.

    Parameters
    ----------
    axes : `~matplotlib.axes` or ``None``
        Axes to plot limb on.
    observer : `astropy.coordinates.SkyCoord`
        Observer coordinate for which the limb is drawn.
    rsun : `~astropy.units.Quantity`
        Solar radius (in physical length units) at which to draw the limb.
        Defaults to the standard photospheric radius.
    resolution : `int`
        The number of points to use to represent the limb.

    Returns
    -------
    visible : `~matplotlib.patches.Polygon` or `~matplotlib.patches.Circle` or None
        The patch added to the axes for the visible part of the limb (i.e., the
        "near" side of the Sun).
    hidden : `~matplotlib.patches.Polygon` or None
        The patch added to the axes for the hidden part of the limb (i.e., the
        "far" side of the Sun).

    Notes
    -----
    Keyword arguments are passed onto the patches.

    If the limb is a true circle, ``visible`` will instead be
    `~matplotlib.patches.Circle` and ``hidden`` will be ``None``.

    If there are no hidden points (e.g., on a synoptic map any limb is fully
    visible) ``hidden`` will be ``None``.

    If there are no visible points (e.g., for an observer on the opposite side
    of the Sun to the map observer) ``visible`` will be ``None``.

    To avoid triggering Matplotlib auto-scaling, these patches are added as
    artists instead of patches. One consequence is that the plot legend is not
    populated automatically when the limb is specified with a text label. See
    :ref:`sphx_glr_gallery_text_labels_and_annotations_custom_legends.py` in
    the Matplotlib documentation for examples of creating a custom legend.
    """
    if not wcsaxes_compat.is_wcsaxes(axes):
        raise ValueError('axes must be a WCSAxes')

    c_kw = {'fill': False,
            'color': 'white',
            'zorder': 100}
    c_kw.update(kwargs)

    transform = axes.get_transform('world')
    # transform is always passed on as a keyword argument
    c_kw.setdefault('transform', transform)

    # If the observer matches the axes's observer frame and is Helioprojective, use a Circle
    axes_frame = wcsapi_to_celestial_frame(axes.wcs)
    if isinstance(axes_frame, Helioprojective):
        axes_observer = SkyCoord(axes_frame.observer)
        if axes_observer.separation_3d(observer) < 1 * u.m:
            distance = observer.transform_to(HeliocentricInertial).spherical.distance
            # Obtain the solar radius and the world->pixel transform
            angular_radius = _angular_radius(rsun, distance)
            circ = patches.Circle([0, 0], radius=angular_radius.to_value(u.deg), **c_kw)
            axes.add_artist(circ)
            return circ, None

    # Otherwise, we use Polygon to be able to distort the limb
    # Create the limb coordinate array using Heliocentric Radial
    limb = get_limb_coordinates(observer, rsun, resolution)

    # Transform the limb to the axes frame and get the 2D vertices
    visible, hidden = _plot_vertices(limb, axes, axes_frame, rsun, **kwargs)

    return visible, hidden


@u.quantity_input
def equator(axes, *, rsun: u.m = R_sun, resolution=500, **kwargs):
    """
    Draws the solar equator as seen by the axes observer.
    Hidden parts are drawn as a dotted line.

    Parameters
    ----------
    axes : `matplotlib.axes.Axes`
        The axes to plot the equator on.
    rsun : `~astropy.units.Quantity`
        Solar radius (in physical length units) at which to draw the solar
        equator. Defaults to the standard photospheric radius.
    resolution : `int`
        The number of points used to represent the equator.

    Returns
    -------
    visible : `~matplotlib.patches.Polygon`
        The patch added to the axes for the visible part of the solar equator.
    hidden : `~matplotlib.patches.Polygon`
        The patch added to the axes for the hidden part of the solar equator.

    Examples
    --------
    >>> import sunpy.map
    >>> from sunpy.visualization import drawing
    >>> import sunpy.data.sample   # doctest: +REMOTE_DATA
    >>> aia = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)   # doctest: +REMOTE_DATA +IGNORE_WARNINGS
    >>> fig = plt.figure()   # doctest: +SKIP
    >>> ax = fig.add_subplot(projection=aia)   # doctest: +SKIP
    >>> aia.plot()   # doctest: +SKIP
    >>> drawing.equator(ax)   # doctest: +SKIP
    """
    if not wcsaxes_compat.is_wcsaxes(axes):
        raise ValueError('axes must be a WCSAxes')
    axes_frame = wcsapi_to_celestial_frame(axes.wcs)

    lat = 0*u.deg
    lat0 = SkyCoord(np.linspace(-180, 179, resolution) * u.deg,
                    np.ones(resolution) * lat, radius=rsun,
                    frame=HeliographicStonyhurst(obstime=axes_frame.obstime))
    visible, hidden = _plot_vertices(lat0, axes, axes_frame, rsun,
                                     **kwargs)
    return visible, hidden


@u.quantity_input
def prime_meridian(axes, *, rsun: u.m = R_sun, resolution=500, **kwargs):
    """
    Draws the solar prime meridian (zero Carrington longitude) as seen by the
    axes observer. Hidden parts are drawn as a dotted line.

    Parameters
    ----------
    axes : `matplotlib.axes.Axes`
        The axes to plot the prime meridian on, or "None" to use current axes.
    rsun : `~astropy.units.Quantity`
        Solar radius (in physical length units) at which to draw the solar
        prime meridian. Defaults to the standard photospheric radius.
    resolution : `int`
        The number of points used to represent the prime meridian.

    Returns
    -------
    visible : `~matplotlib.patches.Polygon`
        The patch added to the axes for the visible part of the solar equator.
    hidden : `~matplotlib.patches.Polygon`
        The patch added to the axes for the hidden part of the solar equator.
    """
    if not wcsaxes_compat.is_wcsaxes(axes):
        raise ValueError('axes must be a WCSAxes')
    axes_frame = wcsapi_to_celestial_frame(axes.wcs)

    if not hasattr(axes_frame, 'observer'):
        raise ValueError('the coordinate frame of the WCSAxes does not have an observer, '
                         'so zero Carrington longitude cannot be determined.')
    observer = axes_frame.observer

    lon = 0*u.deg
    lon0 = SkyCoord(np.ones(resolution) * lon,
                    np.linspace(-90, 90, resolution) * u.deg, radius=rsun,
                    frame=HeliographicCarrington(observer=observer,
                                                 obstime=axes_frame.obstime))
    visible, hidden = _plot_vertices(lon0, axes, axes_frame, rsun,
                                     close_path=False, **kwargs)
    return visible, hidden


@u.quantity_input
def extent(axes, wcs, **kwargs):
    """
    Draws the extent as defined by a given `~astropy.wcs.WCS` instance.

    Parameters
    ----------
    axes : `astropy.visualization.wcsaxes.WCSAxes`
        The axes to plot the WCS extent on, or "None" to use current axes.
    wcs : `~astropy.wcs.wcsapi.BaseLowLevelWCS`
        The WCS that defines the extent to be drawn.

    Returns
    -------
    visible : `~matplotlib.patches.Polygon`
        The patch added to the axes for the visible part of the WCS extent.
    hidden : `~matplotlib.patches.Polygon`
        The patch added to the axes for the hidden part of the WCS extent.
    """
    if not wcsaxes_compat.is_wcsaxes(axes):
        raise ValueError('axes must be a WCSAxes')
    if not isinstance(wcs, BaseHighLevelWCS):
        raise TypeError("wcs should be a High Level WCS object")
    axes_frame = wcsapi_to_celestial_frame(axes.wcs)
    # Traverse the edges of the WCS in pixel space
    edge_pixels = grid_perimeter(*wcs.low_level_wcs.pixel_shape) - 0.5
    edge_coords = wcs.pixel_to_world(*edge_pixels.T)
    # Filter out any non-SkyCoord coordinates returned for these pixel axes
    if not isinstance(edge_coords, SkyCoord):
        edge_coords = [c for c in edge_coords if isinstance(c, SkyCoord)][0]
    visible, hidden = _plot_vertices(edge_coords, axes, axes_frame, R_sun, close_path=True, **kwargs)
    return visible, hidden


def _plot_vertices(coord, axes, frame, rsun, close_path=True, **kwargs):
    """
    Draws the provided SkyCoord on the WCSAxes as `~matplotlib.patches.Polygon`
    objects depending on visibility.
    """
    c_kw = {'fill': False,
            'color': 'white',
            'zorder': 100}
    c_kw.update(kwargs)
    transform = axes.get_transform("world")
    c_kw.setdefault('transform', transform)

    # Get the 2D vertices of the coordinates
    coord = coord.transform_to(frame)
    Tx = coord.spherical.lon.to_value(u.deg)
    Ty = coord.spherical.lat.to_value(u.deg)
    vertices = np.array([Tx, Ty]).T

    # Determine which points are visible (2D points are always visible)
    is_2d = (norm := coord.spherical.norm()).unit is u.one and u.allclose(norm, 1*u.one)
    if not is_2d and hasattr(frame, 'observer'):
        # The reference distance is the distance to the limb for the axes
        # observer
        rsun = getattr(frame, 'rsun', rsun)
        reference_distance = np.sqrt(frame.observer.radius**2 - rsun**2)
        is_visible = coord.spherical.distance <= reference_distance
    else:
        # If the axes has no observer, the entire limb is considered visible
        is_visible = np.ones_like(coord.spherical.distance, bool, subok=False)

    # Identify discontinuities. Uses the same approach as
    # astropy.visualization.wcsaxes.grid_paths.get_lon_lat_path()
    step = np.sqrt((vertices[1:, 0] - vertices[:-1, 0]) ** 2 +
                   (vertices[1:, 1] - vertices[:-1, 1]) ** 2)
    continuous = np.concatenate([[True, True], step[1:] < 100 * step[:-1]])
    if close_path is not True:
        continuous[0] = False
        continuous[-1] = False

    visible, hidden = None, None
    if np.sum(is_visible) > 0:
        if 'linestyle' not in kwargs:
            c_kw['linestyle'] = '-'
        visible = patches.Polygon(vertices, **c_kw)
        _modify_polygon_visibility(visible, is_visible & continuous)
        axes.add_artist(visible)
    if np.sum(~is_visible) > 0:
        if 'linestyle' not in kwargs:
            c_kw['linestyle'] = ':'
        hidden = patches.Polygon(vertices, **c_kw)
        _modify_polygon_visibility(hidden, ~is_visible & continuous)
        axes.add_artist(hidden)

    return visible, hidden


def _modify_polygon_visibility(polygon, keep):
    polygon_codes = polygon.get_path().codes
    polygon_codes[:-1][~keep] = Path.MOVETO
    polygon_codes[-1] = Path.MOVETO if not keep[0] else Path.LINETO



import mmap
from pathlib import Path
from collections import OrderedDict

import numpy as np
import pytest

import astropy.io.fits as fits

from sunpy.data.test import get_test_data_filenames, get_test_filepath
from sunpy.data.test.waveunit import MEDN_IMAGE, MQ_IMAGE, NA_IMAGE, SVSM_IMAGE
from sunpy.io import _fits
from sunpy.util import MetaDict, SunpyMetadataWarning

TEST_RHESSI_IMAGE = get_test_filepath('hsi_image_20101016_191218.fits')
TEST_AIA_IMAGE = get_test_filepath('aia_171_level1.fits')
TEST_EIT_HEADER = get_test_filepath('EIT_header/efz20040301.000010_s.header')
TEST_SWAP_HEADER = get_test_filepath('SWAP/resampled1_swap.header')
TEST_GONG_HEADER = get_test_filepath('gong_synoptic.header')
# Some of the tests images contain an invalid BLANK keyword
# ignore the warning raised by this
pytestmark = pytest.mark.filterwarnings("ignore:Invalid 'BLANK' keyword in header")


@pytest.mark.parametrize(
    ('fname', 'hdus', 'length'),
    [(TEST_RHESSI_IMAGE, None, 4),
     (TEST_RHESSI_IMAGE, 1, 1),
     (TEST_RHESSI_IMAGE, [1, 2], 2),
     (TEST_RHESSI_IMAGE, range(0, 2), 2)]
)
def test_read_hdus(fname, hdus, length):
    pairs = _fits.read(fname, hdus=hdus)
    assert len(pairs) == length


@pytest.mark.parametrize(
    ('fname', 'waveunit'),
    [(TEST_RHESSI_IMAGE, None),
     (TEST_EIT_HEADER, None),
     (TEST_AIA_IMAGE, 'angstrom'),
     (MEDN_IMAGE, 'nm'),
     (MQ_IMAGE, 'angstrom'),
     (NA_IMAGE, 'm'),
     (TEST_SWAP_HEADER, 'angstrom'),
     (SVSM_IMAGE, 'nm'),
     (TEST_GONG_HEADER, 'nm'),]
)
def test_extract_waveunit(fname, waveunit):
    if Path(fname).suffix == '.header':
        header = _fits.format_comments_and_history(fits.Header.fromtextfile(fname))
    else:
        header = _fits.get_header(fname)[0]
    waveunit = _fits.extract_waveunit(header)
    assert waveunit is waveunit


def test_simple_write(tmpdir):
    data, header = _fits.read(TEST_AIA_IMAGE)[0]
    outfile = tmpdir / "test.fits"
    _fits.write(str(outfile), data, header)
    assert outfile.exists()


def test_extra_comment_write(tmpdir):
    data, header = _fits.read(TEST_AIA_IMAGE)[0]
    header["KEYCOMMENTS"]["TEST"] = "Hello world"
    outfile = tmpdir / "test.fits"
    _fits.write(str(outfile), data, header)
    assert outfile.exists()


def test_simple_write_compressed(tmpdir):
    data, header = _fits.read(TEST_AIA_IMAGE)[0]
    outfile = tmpdir / "test.fits"
    _fits.write(str(outfile), data, header, hdu_type=fits.CompImageHDU)
    assert outfile.exists()
    with fits.open(str(outfile)) as hdul:
        assert len(hdul) == 2
        assert isinstance(hdul[1], fits.CompImageHDU)


def test_simple_write_compressed_difftypeinst(tmpdir):
    # `hdu_type=fits.CompImageHDU` and `hdu_type=fits.CompImageHDU()`
    # should produce identical FITS files
    data, header = _fits.read(TEST_AIA_IMAGE)[0]
    outfile_type = str(tmpdir / "test_type.fits")
    outfile_inst = str(tmpdir / "test_inst.fits")
    _fits.write(outfile_type, data, header, hdu_type=fits.CompImageHDU, output_verify="silentfix")
    _fits.write(outfile_inst, data, header, hdu_type=fits.CompImageHDU(), output_verify="silentfix")
    assert fits.FITSDiff(outfile_type, outfile_inst, ignore_comments=["SIMPLE"]).identical


@pytest.mark.parametrize(
    ('kwargs', 'should_fail'),
    [({}, False),
     ({'quantize_level': -32}, True)]
)
def test_simple_write_compressed_instance(tmpdir, kwargs, should_fail):
    data, header = _fits.read(TEST_AIA_IMAGE)[0]
    outfile = tmpdir / "test.fits"

    # Ensure HDU instance is used correctly
    hdu = fits.CompImageHDU(data=np.array([0.]), **kwargs)
    hdu.header['HELLO'] = 'world'  # should be in the written file
    hdu.header['TELESCOP'] = 'other'  # should be replaced with 'SDO/AIA'
    hdu.header['NAXIS'] = 5  # should be replaced with 2
    _fits.write(str(outfile), data, header, hdu_type=hdu, output_verify="silentfix")
    assert outfile.exists()
    with fits.open(str(outfile)) as hdul:
        assert len(hdul) == 2
        assert isinstance(hdul[1], fits.CompImageHDU)
        assert hdul[1].header['HELLO'] == 'world'
        assert hdul[1].header['TELESCOP'] == 'SDO/AIA'
        assert hdul[1].header['NAXIS'] == 2
        data_preserved = hdul[1].data == pytest.approx(data, abs=10)
        print(np.abs(hdul[1].data - data).max())
        print(kwargs)
        if should_fail:  # high compression setting preserved
            assert not data_preserved
        else:
            assert data_preserved


def test_write_with_metadict_header_astropy(tmpdir):
    with fits.open(TEST_AIA_IMAGE) as fits_file:
        data, header = fits_file[0].data, fits_file[0].header
    meta_header = MetaDict(OrderedDict(header))
    temp_file = tmpdir / "temp.fits"
    _fits.write(str(temp_file), data, meta_header)
    assert temp_file.exists()
    fits_file.close()

# Various warnings are thrown in this test, but we just want to check that the code
# works without exceptions


@pytest.mark.filterwarnings('ignore')
def test_fitsheader():
    """Test that all test data can be converted back to a FITS header."""
    extensions = ('.fts', '.fits')
    for ext in extensions:
        bad_file = "not_actually_fits.fits"
        test_files = [f for f in get_test_data_filenames() if f.suffix == ext and f.name != bad_file]
        for ffile in test_files:
            fits_file = fits.open(ffile)
            fits_file.verify("fix")
            meta_header = MetaDict(OrderedDict(fits_file[0].header))
            _fits.header_to_fits(meta_header)
            fits_file.close()


def test_warn_nonascii():
    # Check that a non-ascii character raises a warning and not an error
    with pytest.warns(SunpyMetadataWarning, match='not valid ascii'):
        fits = _fits.header_to_fits({'bad': 'test\t',
                                     'good': 'test'})
    assert 'GOOD' in fits.keys()
    assert 'BAD' not in fits.keys()


def test_warn_nan():
    # Check that a NaN value raises a warning and not an error
    with pytest.warns(SunpyMetadataWarning, match='has a NaN value'):
        fits = _fits.header_to_fits({'bad': float('nan'),
                                     'good': 1.0})
    assert 'GOOD' in fits.keys()
    assert 'BAD' not in fits.keys()


def test_warn_longkey():
    # Check that a key that is too long raises a warning and not an error
    with pytest.warns(SunpyMetadataWarning, match='The meta key badlongkey is too long'):
        fits = _fits.header_to_fits({'badlongkey': 'test',
                                     'goodkey': 'test'})
    assert 'GOODKEY' in fits.keys()
    assert 'BADLONGKEY' not in fits.keys()


def test_read_memmap():
    # This function is a work around for astropy 7.1 vs 7.2
    # In 7.2 there's an extra view on the array returned
    def get_bottom_base(arr):
        if hasattr(arr.base, "base"):
            return get_bottom_base(arr.base)
        return arr.base

    data, _ = _fits.read(TEST_AIA_IMAGE, memmap=True)[0]
    base = get_bottom_base(data)

    # When we are astropy 7.2+ this should be data.base.base
    assert base is not None
    assert isinstance(base, mmap.mmap)

    data, _ = _fits.read(TEST_AIA_IMAGE, memmap=False)[0]
    base = get_bottom_base(data)
    # When we are astropy 7.2+ this should be data.base.base
    assert base is None


def test_merge_multiple_comment_history_entries():
    header = fits.Header()
    history = "First history entry\nSecond history entry"
    comment = "First comment\nSecond comment"
    for hist in history.split("\n"):
        header.add_history(hist)
    for com in comment.split("\n"):
        header.add_comment(com)
    file_header = _fits.format_comments_and_history(header)
    assert file_header["HISTORY"] == history
    assert file_header["COMMENT"] == comment


def test_split_multiline_comment_history_entries():
    header_dict = {
        "history": "First history entry\nSecond history entry",
        "comment": "First comment\nSecond comment",
    }
    fits_header = _fits.header_to_fits(header_dict)
    for k, v in header_dict.items():
        for v_fits, v_dict in zip(fits_header[k.upper()], v.split("\n")):
            assert v_fits == v_dict


def test_roundtrip_comment_history_entries(tmpdir):
    # Construct FITS file from header with multiple history and comment cards
    header = fits.Header.fromtextfile(get_test_filepath("solo_L1_eui-fsi304-image_20201021T145510206_V03.header"))
    data = np.random.rand(header['NAXIS2'], header['NAXIS1'])
    outfilename = str(tmpdir / 'roundtrip-comments-history-astropy.fits')
    fits.writeto(outfilename, data, header)
    # Check roundtripping
    (dh_pair, ) = _fits.read(outfilename)
    header_dict = dh_pair.header
    assert header_dict["HISTORY"] == "\n".join(header["HISTORY"])
    assert header_dict["COMMENT"] == "\n".join(header["COMMENT"])
    outfilename = str(tmpdir / 'roundtrip-comments-history-sunpy.fits')
    _fits.write(outfilename, data, header_dict)
    with fits.open(outfilename) as hdul:
        header_rt = hdul[0].header
    for key in ("HISTORY", "COMMENT"):
        assert len(header_rt[key]) == len(header[key])
        for card_rt, card in zip(header_rt[key], header[key]):
            assert card_rt == card



import collections
from packaging.version import Version
import requests
from datetime import datetime, timedelta
from dateutil.parser import parse as parse_date
from importlib import metadata
from packaging.requirements import Requirement
from functools import cache
import argparse


@cache
def get_package_releases(package):
    """Fetch all versions and their release dates for a package from PyPI"""
    print(f"Querying pypi.org for {package} versions...", end="", flush=True)
    response = requests.get(
        f"https://pypi.org/simple/{package}",
        headers={"Accept": "application/vnd.pypi.simple.v1+json"},
    ).json()
    print("OK")
    file_date = collections.defaultdict(list)
    for f in response["files"]:
        ver = f["filename"].split("-")[1]
        try:
            version = Version(ver)
        except:
            continue
        release_date = None
        for format in ["%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ"]:
            try:
                release_date = datetime.strptime(f["upload-time"], format)
            except ValueError:
                continue
        if not release_date:
            continue
        file_date[version].append(release_date)
    release_date = {v: min(file_date[v]) for v in file_date}
    return release_date


def is_version_old(package, version_str, threshold=timedelta(days=365*2)):
    """Check if a specific version of a package is older than the threshold"""
    releases = get_package_releases(package)
    # Find the release date for the specified version
    target_release_date = None
    for v, date in releases.items():
        if v == Version(version_str):
            target_release_date = date
            break
    if not target_release_date:
        print(f"Did not find version {version_str}")
        return True
    now = datetime.now()
    return (now - target_release_date) > threshold


def find_newest_version(package, threshold=timedelta(days=365*2)):
    """Find the oldest available version that is not older than the threshold"""
    releases = get_package_releases(package)
    # Sort releases by date to easily find the most recent ones
    sorted_releases = dict(sorted(releases.items(), key=lambda x: x[0], reverse=False))
    # Find the first release within the threshold
    for v, date in sorted_releases.items():
        if v.is_prerelease or v.micro != 0:
            continue
        now = datetime.now()
        if (now - date) <= threshold:
            return v, date
    return None, None  # No recent versions found


def get_min_version(requirement):
    """
    Extracts the minimum version from a requirement.

    Parameters:
        requirement (packaging.requirements.Requirement): The requirement object.

    Returns:
        str or None: The minimum version if found, otherwise None.
    """
    spec = requirement.specifier
    # Check for specific version constraints
    versions = []
    # Convert the specifier to a string and split by known operators
    spec_str = str(spec)
    operators = ['==', '>=', '>', '<=', '<', '!=']
    # Iterate over each operator to find matching parts in the spec string
    for op in operators:
        if op in spec_str:
            parts = spec_str.split(op)
            if len(parts) > 1 and not any(o in parts[0] for o in operators):
                versions.append((op, parts[1].strip("'\"")))
    # Determine the minimum version based on the extracted parts
    min_version = None
    for op, ver in versions:
        if op == '==' or (op.startswith(('>=', '>')) and not min_version):
            min_version = ver
    return min_version


def get_requirements(package):
    """
    This wraps `importlib.metadata.requires` to not be garbage.

    Parameters
    ----------
    package : str
        Package you want requirements for.

    Returns
    -------
    `dict`
        A dictionary of requirements with keys being the extra requirement group names.
        The values are a nested dictionary with keys being the package names and
        values being the `packaging.requirements.Requirement` objects.
    """
    requirements: list = metadata.requires(package)
    requires_dict = collections.defaultdict(dict)
    for requirement in requirements:
        req = Requirement(requirement)
        package_name, package_marker = req.name, req.marker
        if package_marker and "extra ==" in str(package_marker):
            group = str(package_marker).split("extra == ")[1].strip('"').strip("'").strip()
        else:
            group = "required"
        # De-duplicate (the same package could appear more than once in the extra == 'all' group)
        if package_name in requires_dict[group]:
            continue
        requires_dict[group][package_name] = req
    return requires_dict


def process_dependencies(package, threshold=timedelta(days=365*2)):
    """
    Processes all dependencies to check their versions against the threshold.

    Parameters:
        package (str): The name of the package.
        requirements_dict (dict): Dictionary returned by get_requirements().

    Returns:
        list: A list of dictionaries containing 'package', 'version',
              'group', and 'is_old' for each dependency.
    """
    requirements_dict = get_requirements(package)
    result = collections.defaultdict(list)
    for group, deps in requirements_dict.items():
        for package_name, req in deps.items():
            min_version = get_min_version(req)
            if min_version:
                is_old = is_version_old(package_name, min_version, threshold=threshold)
                result[group].append({
                    'package': package_name,
                    'version': min_version,
                    'group': group,
                    'is_old': is_old
                })
    return result


def output_version_bumps(package, threshold=timedelta(days=365*2)):
    group_deps = process_dependencies(package)
    for group, deps in group_deps.items():
        print(f"\n{group}\n{'-'*len(group)}")
        for dep in deps:
            if dep['is_old']:
                new_version, release_date = find_newest_version(dep['package'], threshold=threshold)
                if new_version is not None:
                    print(f"{dep['package']} should be bumped to {new_version} which was released on {release_date:%Y-%m-%d}")
                else:
                    print(f"Could not find newer version for {dep['package']}")


def main():
    parser = argparse.ArgumentParser(description='Process package dependancies for minimum bumps.')
    # Required positional argument
    parser.add_argument('--packagename', type=str, default="sunpy", help='The name of the package')
    # Optional integer argument with default value
    parser.add_argument('--threshold', type=int, default=720,
                       help='An optional threshold value in days (default: 720)')
    args = parser.parse_args()
    output_version_bumps(args.packagename, threshold=timedelta(days=args.threshold))


if __name__ == "__main__":
    main()



"""
Updates all the libraries in ``sunpy/extern``
"""

import json
import urllib.request
from pathlib import Path

SUNPY_EXTERN_DIR = Path(__file__).parent.parent / "sunpy" / "extern"

# "filename": ["user", "repository", "path_to_file"]
PACKAGES = {
    "appdirs.py": ["ActiveState", "appdirs", "appdirs.py"],
    "appdirs_license.txt": ["ActiveState", "appdirs", "LICENSE.txt"],
    "distro.py": ["python-distro", "distro", "src/distro/distro.py"],
    "distro_license.rst": ["python-distro", "distro", "LICENSE"],
    # Inflect is not here to avoid a dependency on pydantic.
    # Since it only does one specific thing, we should be able to pin the version permanently.
    "parse.py": ["r1chardj0n3s", "parse", "parse.py"],
    "parse_license.txt": ["r1chardj0n3s", "parse", "LICENSE"],
}


def download_github_file(user: str, repo: str, src: Path, dest: Path):
    """
    Download a file from Github.
    """
    try:
        response = urllib.request.urlopen(f"https://api.github.com/repos/{user}/{repo}")
    except urllib.error.HTTPError as e:
        print(f"Error: {e}")
        return

    url = f"https://api.github.com/repos/{user}/{repo}/tags"
    response = urllib.request.urlopen(url)
    if response.status != 200:
        raise ValueError(f"tags for {user}/{repo} does not exist.")
    response = json.load(response)
    version = response[0]["name"]

    url = f"https://raw.githubusercontent.com/{user}/{repo}/{version}/{src}"
    response = urllib.request.urlopen(url)
    if response.status != 200:
        raise ValueError(f"{url} does not exist.")

    with open(dest, "wb") as f:
        print(f"Updating {dest} to {user}/{repo}:refs/tags/{version}")
        f.write(response.read())


if __name__ == "__main__":
    for package, (user, repo, src) in PACKAGES.items():
        download_github_file(user, repo, src, SUNPY_EXTERN_DIR / package)



"""
SunPy Map

isort:skip_file
"""
# Check if user has installed the map extras
from sunpy.util.sysinfo import _warn_missing_deps
_warn_missing_deps('map')

from sunpy.map.mapbase import *

from sunpy.map import sources
from sunpy.map.header_helper import *
from sunpy.map.map_factory import Map
from sunpy.map.maputils import *
from .compositemap import CompositeMap
from .mapsequence import MapSequence



from unittest.mock import patch

import pytest

from sunpy.util import hash_file
from sunpy.util.exceptions import SunpyUserWarning
from .mocks import MOCK_HASH


def test_cache_basic(cache):
    cache.download('https://example.com/abc.text')
    assert cache._downloader.times_called == 1


def test_cache_caching(cache):
    cache.download('https://example.com/abc.text')
    cache.download('https://example.com/abc.text')
    assert cache._downloader.times_called == 1


def test_cache_redownload(cache):
    cache.download('https://example.com/abc.text')
    cache.download('https://example.com/abc.text', redownload=True)
    assert cache._downloader.times_called == 2


def test_get_by_url(cache):
    cache.download('https://example.com/file_name')
    details = cache._get_by_url('https://example.com/file_name')
    assert details['file_path'].endswith('file_name')


def test_get_by_url_fail(cache):
    cache.download('https://example.com/file_name')
    details = cache._get_by_url('https://example.com/file')
    assert details is None


def test_get_by_hash(cache):
    cache.download('https://example.com/file_name')
    details = cache.get_by_hash(MOCK_HASH)
    assert details['file_path'].endswith('file_name')


def test_get_by_hash_fail(cache):
    cache.download('https://example.com/file_name')
    details = cache.get_by_hash('wrong_hash')
    assert details is None


def test_check_old_file_is_not_removed(cache, mocker):
    # Checks that https://github.com/sunpy/sunpy/issues/7249 is fixed

    # First check that the file is downloaded
    cache.download('https://example.com/file_name')
    first_details = cache.get_by_hash(MOCK_HASH)
    assert first_details['file_path'].endswith('file_name')

    # Force a redownload and check that the old file is not removed
    # We force an error inside `_download_and_hash` to test the full code path
    with patch('sunpy.data.data_manager.cache.hash_file') as download:
        download.side_effect = IOError
        with pytest.warns(SunpyUserWarning, match="Due to the above error, you"):
            path = cache.download('https://example.com/file_name', redownload=True)
        assert download.call_count == 1
    second_details = cache.get_by_hash(MOCK_HASH)
    assert first_details['file_path'] == second_details['file_path'] == str(path)


def test_file_change(cache, mocker):
    cache.download('https://example.com/abc.text')
    file_path = cache.get_by_hash(MOCK_HASH)["file_path"]
    initial_hash = hash_file(file_path)

    with open(file_path, "a") as file:
        file.write("test writing")

    modified_hash = hash_file(file_path)
    assert initial_hash != modified_hash

    with patch('sunpy.data.data_manager.cache.Cache._download_and_hash') as mock_download_and_hash:
        mock_download_and_hash.return_value = (file_path, initial_hash, 'https://example.com/abc.text')
        # Overwrite == True, so file should be overwritten with a new file
        file_path, file_hash, url = mock_download_and_hash('https://example.com/abc.text', redownload=True, overwrite=True)
        assert file_hash == initial_hash

        mock_download_and_hash.return_value = (file_path, modified_hash, "https://example.com/abc.text")
        #Overwrite == False, so the existing file should not be overwritten
        file_path, file_hash, url = mock_download_and_hash('https://example.com/abc.text', redownload=True)
        assert file_hash == modified_hash



from sunpy.data.test import get_test_filepath

MEDN_IMAGE = get_test_filepath('waveunit/medn_halph_fl_20050501_074655.header')
MQ_IMAGE = get_test_filepath('waveunit/mq130812.084253.header')
NA_IMAGE = get_test_filepath('waveunit/na120701.091058.header')
SVSM_IMAGE = get_test_filepath('waveunit/svsm_e3100_S2_20110625_1856.header')



"""
==========================
Masking out the solar disk
==========================

How to mask out all emission from the solar disk.
"""
import matplotlib.pyplot as plt

import sunpy.map
from sunpy.data.sample import AIA_171_IMAGE
from sunpy.map.maputils import all_coordinates_from_map, coordinate_is_on_solar_disk

###############################################################################
# We start with the sample data.

aia = sunpy.map.Map(AIA_171_IMAGE)

###############################################################################
# A utility function gives us access to the helioprojective coordinate of each
# pixels. We can use that to create a new array of all the coordinates
# that are on the solar disk.

hpc_coords = all_coordinates_from_map(aia)

###############################################################################
# Now, we can create a mask from the coordinates by using another utility
# function that gives us a mask that has `True` for those coordinates that are
# on the solar disk. We also make a slight change to the colormap so that
# masked values are shown as black instead of the default white.

mask = coordinate_is_on_solar_disk(hpc_coords)
palette = aia.cmap.copy()
palette.set_bad('black')

###############################################################################
# Finally we create a new map with our new mask.

scaled_map = sunpy.map.Map(aia.data, aia.meta, mask=mask)

###############################################################################
# Let's plot the results using our modified colormap.

fig = plt.figure()
ax = fig.add_subplot(projection=scaled_map)
scaled_map.plot(axes=ax, cmap=palette)
scaled_map.draw_limb(axes=ax)
plt.show()



from datetime import timedelta

import astropy.units as u

from sunpy.net.base_client import QueryResponseTable
from sunpy.time import parse_time


def create_norh_url(date, wavelength):
    year = date.strftime("%Y")
    month = date.strftime("%m")
    day = date.strftime("%d")
    if wavelength == 34 * u.GHz:
        freq = 'tcz'
    elif wavelength == 17 * u.GHz:
        freq = 'tca'
    value_ = year[-2:] + month + day
    base = f"https://solar.nro.nao.ac.jp/norh/data/tcx/{year}/{month}/{freq}{value_}"
    return (base)


def mock_query_object(timerange, client):
    """
    Creating a Query Response object and prefilling it with some information
    """
    # Creating a Query Response Object
    start = timerange.start
    end = timerange.end
    wave = 17*u.GHz
    delta = end-start
    resp = []
    for i in range(delta.datetime.days + 1):
        start_time = start.datetime + timedelta(days=i)
        end_time = start_time + timedelta(days=1) - timedelta(milliseconds=1)
        obj = {
            'Start Time': parse_time(start_time),
            'End Time': parse_time(end_time),
            'Instrument': 'NORH',
            'Source': 'NAOJ',
            'Provider': 'NRO',
            'Wavelength': wave,
            'url': create_norh_url(start_time, wave)
        }
        resp.append(obj)
    results = QueryResponseTable(resp)
    results.client = client
    return results



import re
import tempfile

import pytest
from parfive import Results

import astropy.table
import astropy.time
import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
import sunpy.net.attrs as a
from sunpy.net.jsoc import JSOCClient, JSOCResponse
from sunpy.util.exceptions import SunpyUserWarning

# Ensure all JSOC tests are run on the same parallel worker
pytestmark = [pytest.mark.xdist_group(name="jsoc")]


@pytest.fixture
def client():
    return JSOCClient()


@pytest.fixture
def jsoc_response_double(jsoc_test_email):
    resp = JSOCResponse([{'T_REC': '2011/01/01T00:00', 'INSTRUME': 'AIA'},
                         {'T_REC': '2011/01/02T00:00', 'INSTRUME': 'AIA'}])
    resp.query_args = [{'start_time': astropy.time.Time("2020-01-01T01:00:36.000"),
                        'end_time': astropy.time.Time("2020-01-01T01:00:38.000"),
                        'series': 'hmi.M_45s', 'notify': jsoc_test_email}]
    return resp


def test_jsocresponse_single():
    j1 = JSOCResponse(data=[[1, 2, 3, 4]])
    assert all(j1 == astropy.table.Table(data=[[1, 2, 3, 4]]))
    assert len(j1) == 4


def test_empty_jsoc_response():
    Jresp = JSOCResponse()
    assert len(Jresp) == 0
    assert Jresp.query_args is None
    assert Jresp.requests is None
    assert len(Jresp) == 0


@pytest.mark.remote_data
def test_return_query_args(client):
    res = client.search(a.jsoc.PrimeKey('HARPNUM', 3604),
                        a.jsoc.Series('hmi.sharp_cea_720s'),
                        a.jsoc.Segment('Bp') & a.jsoc.Segment('magnetogram'))
    # Because res.query_args is list that contains dict
    assert 'primekey' in res.query_args[0]


@pytest.mark.remote_data
def test_query(client):
    Jresp = client.search(
        a.Time('2020/1/1T00:00:00', '2020/1/1T00:01:30'),
        a.jsoc.Series('hmi.M_45s'), a.Sample(90 * u.second))
    assert isinstance(Jresp, JSOCResponse)
    assert len(Jresp) == 2


def test_show(client):
    jdict = {'TELESCOP': ['SDO/HMI', 'SDO/AIA'], 'CAR_ROT': [2145, 2145]}
    responses = JSOCResponse(astropy.table.Table(jdict))
    showtable = responses.show('TELESCOP')
    assert isinstance(showtable, astropy.table.Table)
    assert showtable.colnames == ['TELESCOP']
    assert showtable['TELESCOP'][0] == 'SDO/HMI'


@pytest.mark.remote_data
def test_post_notify_fail(client):
    responses = client.search(
        a.Time('2020/1/1T00:00:00', '2020/1/1T00:00:45'),
        a.jsoc.Series('hmi.M_45s'))
    with pytest.raises(ValueError, match="Email address is invalid or not registered"):
        client.request_data(responses)


@pytest.mark.remote_data()
def test_post_wave_series(client):
    with pytest.raises(TypeError, match="The series hmi.M_45s does not support wavelength attribute."):
        client.search(
            a.Time('2020/1/1T00:00:00', '2020/1/1T00:00:45'),
            a.jsoc.Series('hmi.M_45s') | a.jsoc.Series('aia.lev1_euv_12s'),
            a.Wavelength(193 * u.AA) | a.Wavelength(335 * u.AA))


@pytest.mark.remote_data
def test_wait_get(client, jsoc_test_email):
    responses = client.search(
        a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Notify(jsoc_test_email))
    path = tempfile.mkdtemp()
    res = client.fetch(responses, path=path)
    assert isinstance(res, Results)
    assert len(res) == 1


@pytest.mark.remote_data
def test_get_request_tar(client, jsoc_test_email):
    responses = client.search(
        a.Time('2012/1/1T1:00:36', '2012/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Notify(jsoc_test_email))
    bb = client.request_data(responses, method='url-tar')
    bb.wait()
    path = tempfile.mkdtemp()
    aa = client.get_request(bb, path=path)
    assert isinstance(aa, Results)

    responses = client.search(
        a.Time('2012/1/1T1:00:36', '2012/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Notify(jsoc_test_email),
        a.jsoc.Protocol('as-is'))
    bb = client.request_data(responses, method='url-tar')
    bb.wait()
    path = tempfile.mkdtemp()
    aa = client.get_request(bb, path=path)
    assert isinstance(aa, Results)


@pytest.mark.remote_data
def test_invalid_query(client):
    with pytest.raises(ValueError, match="Series must be specified for a JSOC Query"):
        client.search(a.Time('2020/1/1T01:00:00', '2020/1/1T01:00:45'))


@pytest.mark.remote_data
def test_lookup_records_errors(client):
    d1 = {'end_time': astropy.time.Time('2020-01-01 01:00:35'),
          'start_time': astropy.time.Time('2020-01-01 00:00:35')}
    with pytest.raises(ValueError, match="Series must be specified for a JSOC Query"):
        client._lookup_records(d1)

    d1['series'] = 'aia.lev1_euv_12s'
    d1['keys'] = 123
    with pytest.raises(TypeError, match="Keywords can only be passed as a list or comma-separated strings."):
        client._lookup_records(d1)
    d1['keys'] = 'T_OBS'
    d1['primekey'] = {'foo': 'bar'}
    with pytest.raises(ValueError, match=re.escape("Unexpected PrimeKeys were passed. The series aia.lev1_euv_12s supports the following Keywords: ['T_REC', 'WAVELNTH']")):
        client._lookup_records(d1)

    del d1['primekey']
    d1['segment'] = 123
    d1['wavelength'] = 304*u.AA
    with pytest.raises(TypeError, match="Segments can only be passed as a comma-separated string or a list of strings."):
        client._lookup_records(d1)

    d1['segment'] = 'foo'
    with pytest.raises(ValueError, match=re.escape("Unexpected Segments were passed. The series aia.lev1_euv_12s contains the following Segments ['image', 'spikes']")):
        client._lookup_records(d1)

    del d1['segment']
    d1['series'] = 'hmi.m_45s'
    # The series does not support wavelength attribute.
    with pytest.raises(TypeError, match=re.escape("The series hmi.m_45s does not support wavelength attribute. The following primekeys are supported ['T_REC', 'CAMERA']")):
        client._lookup_records(d1)


@pytest.mark.remote_data
def test_make_recordset_errors(client):
    d1 = {'series': 'aia.lev1_euv_12s'}
    with pytest.raises(ValueError, match="Time, Wavelength or an explicit PrimeKey must be specified."):
        client._make_recordset(**d1)

    d1 |= {
        'end_time': astropy.time.Time('2020-01-01 01:00:35', scale='tai'),
        'start_time': astropy.time.Time('2020-01-01 00:00:35', scale='tai'),
        'primekey': {
            'T_REC': '2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI'
        },
    }

    with pytest.raises(ValueError, match=re.escape("Time attribute has been passed both as a Time() and PrimeKey(). Please provide any one of them or separate them by OR operator.")):
        client._make_recordset(**d1)

    d1 |= {
        'end_time': astropy.time.Time('2020-01-01 01:00:35', scale='tai'),
        'start_time': astropy.time.Time('2020-01-01 00:00:35', scale='tai'),
        'wavelength': 604*u.AA,
        'primekey': {'WAVELNTH': '604'},
    }

    with pytest.raises(ValueError, match=re.escape("Wavelength attribute has been passed both as a Wavelength() and PrimeKey(). Please provide any one of them or separate them by OR operator.")):
        client._make_recordset(**d1)


@pytest.mark.remote_data
def test_make_recordset(client):
    d1 = {'series': 'aia.lev1_euv_12s',
          'end_time': astropy.time.Time('2020-01-01 01:00:35', scale='tai'),
          'start_time': astropy.time.Time('2020-01-01 00:00:35', scale='tai')
          }
    exp = 'aia.lev1_euv_12s[2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI]'
    assert client._make_recordset(**d1) == exp

    d1['wavelength'] = 604*u.AA
    exp = 'aia.lev1_euv_12s[2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI][604]'
    assert client._make_recordset(**d1) == exp

    del d1['wavelength']
    d1['primekey'] = ({'WAVELNTH': '604'})
    assert client._make_recordset(**d1) == exp

    del d1['start_time'], d1['end_time']
    d1['primekey'].update({'T_REC': '2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI'})
    exp = 'aia.lev1_euv_12s[2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI]'
    assert client._make_recordset(**d1) == exp

    d1 = {'series': 'hmi.v_45s',
          'end_time': astropy.time.Time('2020-01-01 01:00:35', scale='tai'),
          'start_time': astropy.time.Time('2020-01-01 00:00:35', scale='tai'),
          'segment': 'foo,bar'
          }
    exp = 'hmi.v_45s[2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI]{foo,bar}'
    assert client._make_recordset(**d1) == exp

    d1['segment'] = ['foo', 'bar']
    assert client._make_recordset(**d1) == exp

    d1 = {'series': 'hmi.sharp_720s',
          'end_time': astropy.time.Time('2020-01-01 01:00:35', scale='tai'),
          'start_time': astropy.time.Time('2020-01-01 00:00:35', scale='tai'),
          'segment': ['continuum', 'magnetogram'],
          'primekey': {'HARPNUM': '4864'}
          }
    exp = 'hmi.sharp_720s[4864][2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI]{continuum,magnetogram}'
    assert client._make_recordset(**d1) == exp

    d1['sample'] = 300.0
    exp = 'hmi.sharp_720s[][2020.01.01_00:00:35_TAI-2020.01.01_01:00:35_TAI@300.0s]{continuum,magnetogram}'
    assert client._make_recordset(**d1) == exp


@pytest.mark.remote_data
def test_request_data_error(client, jsoc_test_email):
    responses = client.search(
        a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Notify(jsoc_test_email),
        a.jsoc.Protocol('foo'))
    with pytest.raises(TypeError, match="Protocols other than as-is,fits are not supported."):
        client.request_data(responses)


@pytest.mark.remote_data
def test_request_data_method(client, jsoc_test_email):
    responses = client.search(
        a.Time('2012/1/1T1:00:36', '2012/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Notify(jsoc_test_email))
    req = client.request_data(responses, method='url-tar')
    req.wait()
    assert req.status == 0
    assert req._d['method'] == 'url-tar'
    assert req._d['protocol'] == 'fits'

    responses = client.search(
        a.Time('2012/1/1T1:00:36', '2012/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Notify(jsoc_test_email),
        a.jsoc.Protocol('as-is'))
    req = client.request_data(responses, method='url-tar')
    req.wait()
    assert req.status == 0
    assert req._d['method'] == 'url-tar'
    assert req._d['protocol'] == 'as-is'


def test_can_handle_query_no_series(client):
    assert not client._can_handle_query(a.Time("2020/01/02", "2020/01/03"))
    assert not client._can_handle_query(a.Wavelength(17.1*u.nm))
    assert client._can_handle_query(a.jsoc.Series("hmi.M_45s"))


def test_jsoc_attrs(client):
    attrs = client.load_jsoc_values()
    assert a.jsoc.Series in attrs.keys()
    assert a.jsoc.Segment in attrs.keys()
    assert len(attrs[a.jsoc.Series]) != 0
    assert len(attrs[a.jsoc.Segment]) != 0


@pytest.mark.flaky(reruns_delay=30)
@pytest.mark.remote_data
def test_jsoc_cutout_attrs(client, jsoc_test_email, aia171_test_map):
    cutout = a.jsoc.Cutout(
        SkyCoord(-500*u.arcsec, -275*u.arcsec, frame=aia171_test_map.coordinate_frame),
        top_right=SkyCoord(150*u.arcsec, 375*u.arcsec, frame=aia171_test_map.coordinate_frame),
        tracking=True
    )
    q = client.search(
        a.Time(aia171_test_map.date, aia171_test_map.date + 1 * u.min),
        a.Wavelength(171*u.angstrom),
        a.jsoc.Series.aia_lev1_euv_12s,
        a.jsoc.Notify(jsoc_test_email),
        a.jsoc.Segment.image,
        cutout,
    )
    req = client.request_data(q, method='url', protocol='fits')
    req.wait()
    assert req.status == 0
    files = client.get_request(req)
    assert len(files) == 6
    m = sunpy.map.Map(files, sequence=True)
    assert m.data.shape == (1085, 1085, 6)


def test_row_and_warning(mocker, client, jsoc_response_double):
    mocker.patch("sunpy.net.jsoc.jsoc.JSOCClient.get_request")
    request_data = mocker.patch("sunpy.net.jsoc.jsoc.JSOCClient.request_data")
    with pytest.warns(SunpyUserWarning):
        client.fetch(jsoc_response_double[0], sleep=0)
    request_data.assert_called_once_with(jsoc_response_double[0].as_table())


@pytest.mark.remote_data
def test_check_request_keywords(client):
    responses = client.search(
        a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Keyword("QUALITY") == 1)
    assert len(responses) == 0

    responses = client.search(
        a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Keyword("QUALITY") == 0)
    assert len(responses) == 1

    responses = client.search(
        a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
        a.jsoc.Series('hmi.M_45s'), a.jsoc.Keyword("QUALITY") < 2)
    assert len(responses) == 1

    with pytest.raises(ValueError, match="Keyword: 'EXPTIME' is not supported by series:"):
        client.search(
            a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
            a.jsoc.Series('hmi.M_45s'), a.jsoc.Keyword("EXPTIME") < 2)

    responses = client.search(
        a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
        a.jsoc.Series('aia.lev1_euv_12s'), a.jsoc.Keyword("QUALITY") < 2, a.jsoc.Keyword("EXPTIME") < 2)
    assert len(responses) == 0

    responses = client.search(
        a.Time('2020/1/1T1:00:36', '2020/1/1T01:00:38'),
        a.jsoc.Series('aia.lev1_euv_12s'), a.jsoc.Keyword("QUALITY") < 2, a.jsoc.Keyword("EXPTIME") > 2)
    assert len(responses) == 7


def test_empty_response_fetch(client):
    # Check that results from downloading an empty response are empty
    #
    # Could do a search like below, but to save a JSOC query in the tests
    # just test against an empty response.
    #
    # response = client.search(
    #     a.Time('1990-12-12T00:00:00', '1990-12-12T0:01:00'),
    #    a.jsoc.Series.hmi_b_720s
    # )
    response = JSOCResponse()
    assert len(response) == 0
    result = client.fetch(response)
    assert len(result) == 0


@pytest.mark.remote_data
def test_string_keyword_is_converted_correctly(client):
    # Bug reported here:
    # https://community.openastronomy.org/t/how-to-select-only-fuv-iris-level-1-data-using-jsoc/900
    responses = client.search(
        a.Time("2014-01-01T00:00", "2014-01-01T00:05"),
        a.jsoc.Series("iris.lev1"), a.jsoc.Keyword("INSTRUME") == "FUV"
    )
    assert len(responses) == 6

@pytest.mark.remote_data
def test_segments_query(client):
    res = client.search(a.jsoc.PrimeKey('HARPNUM', 3604),
                        a.jsoc.Series('hmi.sharp_cea_720s'))
    seg_res = client.search(a.jsoc.PrimeKey('HARPNUM', 3604),
                        a.jsoc.Series('hmi.sharp_cea_720s'),
                        a.jsoc.Segment('Bp') & a.jsoc.Segment('magnetogram'))
    # Check that the correct segments are returned and are not returned in the original query
    assert len(res) == len(seg_res) == 1223
    assert len(res.columns) != len(seg_res.columns)
    assert len(seg_res.columns) - len(res.columns) == 2
    assert "magnetogram" not in res.keys()
    assert "Bp" not in res.keys()
    assert "magnetogram" in seg_res.keys()
    assert "Bp" in seg_res.keys()
    # SUMS directories are like /SUM93/D1044682184/S00000/Bp.fits
    # So checking the start and end is the easiest way to check the segments
    assert seg_res["magnetogram"][0].startswith("/SUM")
    assert seg_res["Bp"][0].startswith("/SUM")
    assert seg_res["magnetogram"][0].endswith("magnetogram.fits")
    assert seg_res["Bp"][0].endswith("Bp.fits")



import numpy as np
import pytest
from matplotlib.figure import Figure

from astropy.coordinates.matrix_utilities import rotation_matrix

from sunpy.data.test import get_test_filepath
from sunpy.image.transform import _rotation_registry, affine_transform
from sunpy.tests.helpers import figure_test, skip_opencv, skip_skimage
from sunpy.util import SunpyUserWarning

_methods_to_marks = {"scikit-image": skip_skimage, "opencv": skip_opencv}
_rotation_registry_params = [pytest.param(meth, marks=_methods_to_marks.get(meth, [])) for meth in _rotation_registry]

# Tolerance for tests
RTOL = 1.0e-10


@pytest.fixture
def original():
    filepath = get_test_filepath("camera.npy")
    return np.load(filepath).astype('float')


@pytest.fixture
def identity():
    return np.array([[1, 0], [0, 1]])


@pytest.fixture
def rot30():
    return rotation_matrix(30)[0:2, 0:2]


def compare_results(expect, result, allclose=True):
    """
    Function to check that the obtained results are what was expected, to
    within the relative tolerance defined above.
    """
    # Outermost pixels can contain artefacts which will be ignored.
    exp = expect[1:-1, 1:-1]
    res = result[1:-1, 1:-1]
    t1 = abs(exp.mean() - res.mean()) <= RTOL*exp.mean()

    # Don't do the allclose test for skimage due to its forced interpolation beyond the edge of the
    # original image
    if not allclose:
        return t1
    else:
        notclose = ~np.isclose(exp, res, rtol=RTOL)
        t2 = not np.any(notclose)

        # Print out every mismatch
        if not t2:
            with np.errstate(divide='ignore'):
                mismatches = np.stack([*notclose.nonzero(), exp[notclose], res[notclose]]).T
                for row in mismatches:
                    print(f"i={int(row[0]+1)}, j={int(row[1]+1)}: ",
                          f"expected={row[2]}, result={row[3]}, "
                          f"adiff={row[2]-row[3]}, rdiff={(row[2]-row[3])/row[2]}")

    return t1 and t2


@pytest.mark.parametrize(('angle', 'k'), [(90.0, 1), (-90.0, -1), (-270.0, 1),
                                          (-90.0, 3), (360.0, 0), (-360.0, 0)])
def test_rotation(original, angle, k):
    # Test rotation against expected outcome
    angle = np.radians(angle)
    c = np.round(np.cos(angle))
    s = np.round(np.sin(angle))
    rmatrix = np.array([[c, -s], [s, c]])
    expected = np.rot90(original, k=k)

    # Run the tests at order 4 as it produces more accurate 90 deg rotations
    rot = affine_transform(original, order=4, rmatrix=rmatrix)
    assert compare_results(expected, rot)

    # TODO: Check incremental 360 degree rotation against original image

    # Check derotated image against original
    derot_matrix = np.array([[c, s], [-s, c]])
    derot = affine_transform(rot, order=4, rmatrix=derot_matrix)
    assert compare_results(original, derot)


@pytest.mark.parametrize(('angle', 'k'), [(90.0, 1), (-90.0, -1), (-270.0, 1),
                                          (-90.0, 3), (360.0, 0), (-360.0, 0)])
def test_skimage_rotation(original, angle, k):
    pytest.importorskip("skimage")
    # Test rotation against expected outcome
    angle = np.radians(angle)
    c = np.round(np.cos(angle))
    s = np.round(np.sin(angle))
    rmatrix = np.array([[c, -s], [s, c]])
    expected = np.rot90(original, k=k)
    rot = affine_transform(original, rmatrix=rmatrix, method='scikit-image')
    assert compare_results(expected, rot, allclose=False)

    # TODO: Check incremental 360 degree rotation against original image

    # Check derotated image against original
    derot_matrix = np.array([[c, s], [-s, c]])
    derot = affine_transform(rot, rmatrix=derot_matrix, method='scikit-image')
    assert compare_results(original, derot, allclose=False)


dx_values, dy_values = list(range(-100, 101, 100))*3, list(range(-100, 101, 100))*3
dy_values.sort()


@pytest.mark.parametrize(('dx', 'dy'), list(zip(dx_values, dy_values)))
def test_shift(original, dx, dy):
    # Rotation center for all translation tests.
    image_center = np.array(original.shape)/2.0 - 0.5

    # No rotation for all translation tests.
    rmatrix = np.array([[1.0, 0.0], [0.0, 1.0]])

    # Check a shifted shape against expected outcome
    expected = np.roll(np.roll(original, dx, axis=1), dy, axis=0)
    rcen = image_center - np.array([dx, dy])
    shift = affine_transform(original, rmatrix=rmatrix, recenter=True, image_center=rcen, missing=0)
    ymin, ymax = max([0, dy]), min([original.shape[1], original.shape[1]+dy])
    xmin, xmax = max([0, dx]), min([original.shape[0], original.shape[0]+dx])
    assert compare_results(expected[ymin:ymax, xmin:xmax], shift[ymin:ymax, xmin:xmax])

    # Check shifted and unshifted shape against original image
    rcen = image_center + np.array([dx, dy])
    unshift = affine_transform(shift, rmatrix=rmatrix, recenter=True, image_center=rcen, missing=0)
    # Need to ignore the portion of the image cut off by the first shift
    ymin, ymax = max([0, -dy]), min([original.shape[1], original.shape[1]-dy])
    xmin, xmax = max([0, -dx]), min([original.shape[0], original.shape[0]-dx])
    assert compare_results(original[ymin:ymax, xmin:xmax], unshift[ymin:ymax, xmin:xmax])


@pytest.mark.parametrize("scale_factor", [0.25, 0.5, 0.75, 1.0, 1.25, 1.5])
def test_scale(original, scale_factor):
    tf = pytest.importorskip("skimage.transform")
    # No rotation for all scaling tests.
    rmatrix = np.array([[1.0, 0.0], [0.0, 1.0]])

    # Check a scaled image against the expected outcome
    # When we depend on SciPy 1.6, we can replace this with scipy.ndimage.zoom(..., grid_mode=True)
    newim = tf.rescale(original / original.max(), scale_factor, order=1,
                       mode='constant', anti_aliasing=False) * original.max()
    # Old width and new center of image
    w = original.shape[0] / 2.0 - 0.5
    new_c = (newim.shape[0] / 2.0) - 0.5
    expected = np.zeros(original.shape)
    upper = int(w + new_c + 1)
    if scale_factor > 1:
        lower = int(new_c - w)
        expected = newim[lower:upper, lower:upper]
    else:
        lower = int(w - new_c)
        expected[lower:upper, lower:upper] = newim
    scale = affine_transform(original, rmatrix=rmatrix, scale=scale_factor, order=1, missing=0)
    assert compare_results(expected, scale)


@pytest.mark.parametrize(('angle', 'dx', 'dy', 'scale_factor'), [(90, -100, 40, 0.25),
                                                                 (-90, 40, -80, 0.75),
                                                                 (180, 20, 50, 1.5)])
def test_all(original, angle, dx, dy, scale_factor):
    """
    Tests to make sure that combinations of scaling, shifting and rotation
    produce the expected output.
    """
    tf = pytest.importorskip("skimage.transform")
    k = int(angle / 90)
    angle = np.radians(angle)
    image_center = np.array(original.shape) / 2.0 - 0.5

    # Check a shifted, rotated and scaled shape against expected outcome
    c = np.round(np.cos(angle))
    s = np.round(np.sin(angle))
    rmatrix = np.array([[c, -s], [s, c]])
    # When we depend on SciPy 1.6, we can replace this with scipy.ndimage.zoom(..., grid_mode=True)
    scale = tf.rescale(original / original.max(), scale_factor, order=1,
                       mode='constant', anti_aliasing=False) * original.max()
    new = np.zeros(original.shape)

    disp = np.array([dx, dy])
    dxs, dys = np.asarray(disp * scale_factor, dtype=int)
    # Old width and new center of image
    w = np.array(original.shape[0])/2.0 - 0.5
    new_c = (np.array(scale.shape[0])/2.0 - 0.5)
    upper = int(w+new_c+1)
    if scale_factor > 1:
        lower = int(new_c-w)
        new = scale[lower-dys:upper-dys, lower-dxs:upper-dxs]
    else:
        lower = int(w-new_c)
        new[lower+dys:upper+dys, lower+dxs:upper+dxs] = scale
    rcen = image_center - disp
    expected = np.rot90(new, k=k)

    rotscaleshift = affine_transform(original, rmatrix=rmatrix, scale=scale_factor, order=1,
                                     recenter=True, image_center=rcen, missing=0)
    assert compare_results(expected, rotscaleshift)

    # Check a rotated/shifted and restored image against original
    transformed = affine_transform(original, rmatrix=rmatrix, scale=1.0, order=1, recenter=True,
                                   image_center=rcen, missing=0)
    inv_rcen = image_center + np.dot(rmatrix.T, np.array([dx, dy]))
    inverse = affine_transform(transformed, rmatrix=rmatrix.T, scale=1.0, order=1, recenter=True,
                               image_center=inv_rcen, missing=0)

    # Need to ignore the portion of the image cut off by the first shift
    ymin, ymax = max([0, -dy]), min([original.shape[1], original.shape[1]-dy])
    xmin, xmax = max([0, -dx]), min([original.shape[0], original.shape[0]-dx])
    assert compare_results(original[ymin:ymax, xmin:xmax], inverse[ymin:ymax, xmin:xmax])


def test_flat(identity):
    # Test that a flat array can be rotated using scikit-image
    in_arr = np.array([[100]], dtype=np.float64)
    out_arr = affine_transform(in_arr, rmatrix=identity)
    assert np.allclose(in_arr, out_arr, rtol=RTOL)


def test_nan_skimage(identity):
    pytest.importorskip("skimage")
    # Test preservation of NaN values for scikit-image rotation
    in_arr = np.array([[np.nan, 0]])
    out_arr = affine_transform(in_arr, rmatrix=identity, order=0, method='scikit-image')
    assert np.isnan(out_arr[0, 0])


def test_nan_scipy(identity):
    # Test preservation of NaN values for scipy rotation
    in_arr = np.array([[np.nan, 0]])
    out_arr = affine_transform(in_arr, rmatrix=identity, order=0, method='scipy')
    assert np.isnan(out_arr[0, 0])


def test_int(identity):
    pytest.importorskip("skimage")
    # Test casting of integer array to float array
    in_arr = np.array([[100]], dtype=int)
    with pytest.warns(SunpyUserWarning, match='Integer input data has been cast to float64'):
        out_arr = affine_transform(in_arr, rmatrix=identity, method='scikit-image')
    assert np.issubdtype(out_arr.dtype, np.floating)


def test_float32(identity):
    # Check that float32 input remains as float32 output
    # Test casting of integer array to float array
    in_arr = np.array([[100]], dtype=np.float32)
    out_arr = affine_transform(in_arr, rmatrix=identity)
    assert np.issubdtype(out_arr.dtype, np.float32)


def test_reproducible_matrix_multiplication():
    # Test whether matrix multiplication involving a large matrix always gives the same answer
    # This indirectly tests whichever BLAS/LAPACK libraries that NumPy is linking to (if any)
    x = np.arange(500000, dtype=np.float64)
    src = np.vstack((x, -10*x)).T
    matrix = np.array([[0, 1], [1, 0]])

    expected = np.vstack((-10*x, x)).T  # src @ matrix

    mismatches = np.zeros(500, int)
    for i in range(len(mismatches)):
        result = src @ matrix
        mismatches[i] = (~np.isclose(result, expected)).sum()
        if mismatches[i] != 0:
            print(f"{mismatches[i]} mismatching elements in multiplication #{i}")

    assert np.sum(mismatches != 0) == 0


@figure_test
def test_clipping(rot30):
    # Generates a plot to test the clipping the output image to the range of the input image
    image = np.ones((20, 20))
    image[4:-4, 4:-4] = 2

    num_methods = len(_rotation_registry.keys())

    fig = Figure(figsize=(12, 2*num_methods))
    axs = fig.subplots(nrows=num_methods, ncols=5)

    for i, method in enumerate(_rotation_registry.keys()):
        axs[i, 0].imshow(image, vmin=0, vmax=3)
        axs[i, 1].imshow(affine_transform(image, rot30, clip=False, method=method, missing=0),
                         vmin=0, vmax=3)
        axs[i, 2].imshow(affine_transform(image, rot30, clip=True, method=method, missing=0),
                         vmin=0, vmax=3)
        axs[i, 3].imshow(affine_transform(image, rot30, clip=True, method=method, missing=2),
                         vmin=0, vmax=3)
        axs[i, 4].imshow(affine_transform(image, rot30, clip=True, method=method, missing=np.nan),
                         vmin=0, vmax=3)
        axs[i, 0].set_ylabel(method)

    axs[0, 0].set_title('Original')
    axs[0, 1].set_title('no clip & missing=0')
    axs[0, 2].set_title('clip & missing=0')
    axs[0, 3].set_title('clip & missing=2')
    axs[0, 4].set_title('clip & missing=NaN')

    return fig


@pytest.mark.filterwarnings("ignore:.*bug in the implementation of scikit-image")
@figure_test
def test_nans(rot30):
    # Generates a plot to test the preservation and expansions of NaNs by the rotation
    image_with_nans = np.ones((23, 23))
    image_with_nans[4:-4, 4:-4] = 2
    image_with_nans[9:-9, 9:-9] = np.nan

    num_methods = len(_rotation_registry.keys())

    fig = Figure(figsize=(16, 2*num_methods))
    axs = fig.subplots(nrows=num_methods, ncols=7)

    axs[0, 0].set_title('Original (NaNs are white)')

    for j in range(6):
        axs[0, j+1].set_title(f'order={j}')
    for i, method in enumerate(_rotation_registry.keys()):
        axs[i, 0].imshow(image_with_nans, vmin=-1.1, vmax=1.1)
        for j in range(6):
            if j not in _rotation_registry[method].allowed_orders:
                with pytest.raises(ValueError, match=f"^{j} is one of the allowed orders for method '{method}': {set(_rotation_registry[method].allowed_orders)}$"):
                    affine_transform(image_with_nans, rot30, order=j, method=method, missing=np.nan)
                axs[i, j+1].remove()
            else:
                axs[i, j+1].imshow(affine_transform(image_with_nans, rot30,
                                                    order=j, method=method, missing=np.nan),
                                   vmin=-1.1, vmax=1.1)
        axs[i, 0].set_ylabel(method)

    return fig


@pytest.mark.filterwarnings("ignore:.*bug in the implementation of scikit-image")
@pytest.mark.parametrize('method', _rotation_registry_params)
@pytest.mark.parametrize('order', range(6))
def test_endian(method, order, rot30):
    if order not in _rotation_registry[method].allowed_orders:
        return

    # Test that the rotation output values do not change with input byte order
    native = np.ones((10, 10))
    swapped = native.view(native.dtype.newbyteorder()).byteswap()

    rot_native = affine_transform(native, rot30, order=order, method=method, missing=0)
    rot_swapped = affine_transform(swapped, rot30, order=order, method=method, missing=0)

    assert compare_results(rot_native, rot_swapped)



"""Test cases for STEREO Map subclasses.
This particular test file pertains to HIMap.
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.stereo import HIMap
from .helpers import _test_private_date_setters


@pytest.fixture
def hi_map():
    return get_dummy_map_from_header(get_test_filepath('hi_20110910_114721_s7h2A.header'))


def test_fitstoHI(hi_map):
    """Tests the creation of HIMap to fits"""
    assert isinstance(hi_map, HIMap)


def test_is_datasource_for(hi_map):
    """Test the is_data_source_for method of HIMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert hi_map.is_datasource_for(hi_map.data, hi_map.meta)


def test_reference_date(hi_map):
    assert hi_map.reference_date.isot == "2011-09-10T11:47:46.004"


def test_date(hi_map):
    assert hi_map.date.isot == "2011-09-10T11:47:21.005"


def test_private_date_setters(hi_map):
    _test_private_date_setters(hi_map)


def test_measurement(hi_map):
    """Tests the measurement property of the HIMap object."""
    assert hi_map.measurement == "white-light"


def test_observatory(hi_map):
    """Tests the observatory property of the HIMap object."""
    assert hi_map.observatory == "STEREO A"


def test_norm_clip(hi_map):
    # Tests that the default normalizer has clipping disabled
    assert not hi_map.plot_settings['norm'].clip


def test_wcs(hi_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    hi_map.pixel_to_world(0*u.pix, 0*u.pix)



import io
import re

import pytest

from sunpy.tests.mocks import MockHTTPResponse, MockObject, MockOpenTextFile


@pytest.fixture
def mocked_mockobject():
    return MockObject(records=12)


def test_MockObject_illegal_kwargs(mocked_mockobject):
    """
    Any attempt to use a kwarg which has the same name as an attribute/method
    of the underlying object or datastore will raise a ValueError.
    """
    with pytest.raises(ValueError, match="kwarg 'values' is already an attribute"):
        MockObject(records=[], values=1)

    with pytest.raises(ValueError, match="kwarg 'items' is already an attribute"):
        MockObject(items=('a', 'b', 'c'))

    with pytest.raises(ValueError, match="kwarg '__hash__' is already an attribute"):
        MockObject(__hash__=0x23424)

    # adding a new 'prohibited' attribute will be prevented
    with pytest.raises(ValueError, match="Name 'keys' is already an attribute"):
        mocked_mockobject['keys'] = [3, 4]


def test_MockObject_attr(mocked_mockobject):
    """
    builtin hasattr & getattr functions, these don't work on dictionaries but
    they do on classes.
    """
    assert hasattr(mocked_mockobject, 'records') is True
    assert hasattr(mocked_mockobject, 'cost') is False

    assert getattr(mocked_mockobject, 'records') == 12

    with pytest.raises(AttributeError):
        getattr(mocked_mockobject, 'jobs')


def test_MockObject_get(mocked_mockobject):
    """
    Getting attributes from `MockObject` using dot and bracket notation.
    """
    assert mocked_mockobject['records'] == 12
    assert mocked_mockobject.records == 12

    with pytest.raises(AttributeError):
        mocked_mockobject.no_key

    with pytest.raises(KeyError):
        mocked_mockobject['not-here']


def test_MockObject_set_get(mocked_mockobject):
    """
    Setting attributes in `MockObject` using bracket notation *not* dot
    notation.
    """

    # Only change the value of existing & new items using 'bracket' notation
    mocked_mockobject['records'] = 45
    assert mocked_mockobject.records == 45
    assert mocked_mockobject['records'] == 45

    # Using 'dot' notation will set a new attribute on 'MockObject' not on the datastore
    # DO NOT DO THIS!
    mocked_mockobject.records = -344

    # This is equivalent to seattr(mocked_mockobject, 'records', -344). Again, don't do this!
    assert mocked_mockobject.records == -344

    # The 'real' value remains unchanged.
    assert mocked_mockobject['records'] == 45


def test_MockObject_len():
    """
    Testing ``MockObject.__len__``.
    """
    assert len(MockObject(responses=['a', 'b', 'c', 'd'], requests=(1, 2, 3))) == 2


def test_MockObject_del(mocked_mockobject):
    """
    Ensure ``MockObject.__delitem__`` is **not** implemented.
    """
    with pytest.raises(NotImplementedError):
        del mocked_mockobject['records']


def test_MockObject_iter(mocked_mockobject):
    """
    Test ``MockObject.__iter__``.
    """
    assert list(iter(mocked_mockobject)) == ['records']


def test_repr_MockObject():
    """
    Test ``MockObject.__repr__``.
    """
    empty = MockObject()

    mo_p = re.compile(r"^(?P<_><)sunpy\.tests\.mocks\.MockObject \{\} "
                      "at 0x[0-9A-Fa-f]+L?(?(_)>|)$")
    assert mo_p.match(repr(empty)) is not None


def test_read_only_mode_MockOpenTextFile():
    """
    Reading from a read only file, writing should be prohibited.
    """
    new_line = '\n'
    content = fr'a{new_line}bc{new_line}nd{new_line}{new_line}'

    read_only = MockOpenTextFile('rom.txt', data=content)
    assert read_only.readable() is True
    assert read_only.writable() is False

    with pytest.raises(io.UnsupportedOperation):
        read_only.write('')

    assert read_only.read() == content
    assert read_only.readlines() == [f'{line}{new_line}'
                                     for line in content.split(new_line)]
    read_only.close()

    with pytest.raises(ValueError, match="I/O operation on closed file"):
        read_only.readable()

    with pytest.raises(ValueError, match="I/O operation on closed file"):
        read_only.writable()

    with pytest.raises(ValueError, match="I/O operation on closed file"):
        read_only.read()

    with pytest.raises(ValueError, match="I/O operation on closed file"):
        read_only.readlines()


def test_write_only_mode_MockOpenTextFile():
    """
    Writing to to write-only file, reading should be prohibited.
    """
    write_only = MockOpenTextFile('write.txt', 'w')

    assert write_only.readable() is False
    assert write_only.writable() is True

    with pytest.raises(io.UnsupportedOperation):
        write_only.read()

    data = '0123456789'

    num_chars = write_only.write(data)
    assert num_chars == len(data)


def test_read_and_write_MockOpenTextFile():
    """
    Reading & writing to a file with read/write access.
    """
    rd_wr = MockOpenTextFile(mode='r+')

    assert rd_wr.name == 'N/A'
    assert rd_wr.readable() is True
    assert rd_wr.writable() is True

    # Initially empty
    assert rd_wr.read() == ''

    data = '0123456789'

    num_chars = rd_wr.write(data)
    assert num_chars == len(data)

    assert rd_wr.read() == data

    rd_wr.close()


def test_repr_MockOpenTextFile():
    """
    Test ``MockOpenTextFile.__repr__``.
    """
    mo_p = re.compile(
        r"^(?P<_><)sunpy\.tests\.mocks\.MockOpenTextFile file \'a\' "
        "mode \'r\' at 0x[0-9A-Fa-f]+L?(?(_)>|)$"
    )
    assert mo_p.match(repr(MockOpenTextFile('a', 'r'))) is not None


def test_MockHTTPResponse():
    """
    Simple tests querying the headers attribute.
    """
    headers = {
        'Content-Type': 'text/html',
        'Content-Disposition': 'attachment; filename="filename.jpg"'
    }
    response = MockHTTPResponse(url='https://abc.com', headers=headers)
    assert response.url == 'https://abc.com'
    assert response.headers.get('Content-Disposition') == 'attachment; filename="filename.jpg"'
    assert response.headers.get('Content-Length') is None
    # Key *not* case insensitive
    assert response.headers.get('content-type') is None




import pytest

import sunpy.timeseries
from sunpy.data.test import get_test_filepath
from sunpy.tests.helpers import figure_test
from sunpy.util.exceptions import SunpyUserWarning

esp_filepath = get_test_filepath('eve_l1_esp_2011046_00_truncated.fits')
eve_filepath = get_test_filepath('EVE_L0CS_DIODES_1m_truncated.txt')


@pytest.mark.skipif(pytest.__version__ < "8.0.0", reason="pytest >= 8.0.0 raises an extra warning for this test")
def test_eve():
    with pytest.warns(SunpyUserWarning, match='Unknown units for oldXRSB proxy'):
        with pytest.warns(SunpyUserWarning, match='Unknown units for x_cool'):
            ts_eve = sunpy.timeseries.TimeSeries(eve_filepath, source='EVE')
    assert isinstance(ts_eve, sunpy.timeseries.sources.eve.EVESpWxTimeSeries)


def test_implicit_esp():
    # Test an ESP TimeSeries
    ts_esp = sunpy.timeseries.TimeSeries(esp_filepath)
    assert isinstance(ts_esp, sunpy.timeseries.sources.eve.ESPTimeSeries)


def test_esp():
    # Test an ESP TimeSeries
    ts_esp = sunpy.timeseries.TimeSeries(esp_filepath, source='ESP')
    assert isinstance(ts_esp, sunpy.timeseries.sources.eve.ESPTimeSeries)


def test_esp_plot_column(esp_test_ts):
    axes = esp_test_ts.plot(columns=['QD', 'CH_18', 'CH_36'])
    assert len(axes) == 3
    assert '0.1-7nm' in axes[0].get_ylabel()
    assert '18nm' in axes[1].get_ylabel()
    assert '36nm' in axes[2].get_ylabel()


@figure_test
def test_esp_peek(esp_test_ts):
    esp_test_ts.peek()


@figure_test
def test_eve_peek(eve_test_ts):
    eve_test_ts.peek()



import pytest

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.net.jsoc as jsoc
import sunpy.net.jsoc.attrs as attrs
from sunpy.coordinates import frames
from sunpy.net import _attrs as core_attrs
from sunpy.net.attr import AttrAnd, AttrOr


@pytest.mark.parametrize((("attr1", "attr2")),
                         [(attrs.Series('foo'), attrs.Series('boo')),
                          (attrs.Protocol('a1'), attrs.Protocol('a2')),
                          (attrs.Notify('email@somemail.com'),
                           attrs.Notify('someemail@somemail.com'))])
def test_and(attr1, attr2):
    with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for \&"):
        attr1 & attr2


def test_basicquery():
    a1 = attrs.Series('foo')
    t1 = core_attrs.Time('2012/01/01', '2013/1/2')
    ans1 = jsoc.jsoc.and_(a1, t1)
    assert isinstance(ans1, AttrAnd)
    assert len(ans1.attrs) == 2


def test_mediumquery():
    a1 = attrs.Series('foo1')
    a2 = attrs.Series('foo2')
    t1 = core_attrs.Time('2012/01/01', '2013/1/2')
    ans1 = jsoc.jsoc.and_(a1 | a2, t1)
    assert isinstance(ans1, AttrOr)
    assert isinstance(ans1.attrs[0], AttrAnd)
    assert isinstance(ans1.attrs[1], AttrAnd)


def test_complexquery():
    a1 = attrs.Series('foo1')
    a2 = attrs.Series('foo2')
    t1 = core_attrs.Time('2012/01/01', '2013/1/2')
    t2 = core_attrs.Time('2012/01/01', '2013/1/3')
    ans1 = jsoc.jsoc.and_(a1 | a2, t1 | t2)
    assert isinstance(ans1.attrs[0], AttrOr)
    assert isinstance(ans1.attrs[0].attrs[0], AttrAnd)
    assert isinstance(ans1.attrs[0].attrs[1], AttrAnd)


def test_wavelength_error():
    with pytest.raises(TypeError):
        attrs.Wavelength('wobble')
    with pytest.raises(TypeError):
        attrs.Wavelength(3.24)
    with pytest.raises(TypeError):
        attrs.Wavelength((3, 3))


def test_wave_self():
    w1 = attrs.Wavelength(193*u.AA)
    assert jsoc.jsoc.and_(w1 | w1) is w1


def test_duplicate():
    w1 = attrs.Wavelength(193*u.AA)
    w2 = attrs.Wavelength(193*u.AA)
    assert jsoc.jsoc.and_(w1 | w2).min is w1.min


def test_random():
    w1 = attrs.Wavelength(193*u.AA)
    w2 = attrs.Series('spam')
    assert jsoc.jsoc.and_(w1 | w2) == AttrOr([w1, w2])


def test_empty_notify():
    with pytest.raises(ValueError, match="Notify attribute must contain an email address"):
        attrs.Notify(None)


def test_not_email_notify():
    with pytest.raises(ValueError, match="Notify attribute must contain an '@' symbol to be a valid email address"):
        attrs.Notify("someemailthatisntone")


def test_cutout_not_helioprojective():
    hpc = SkyCoord(500*u.arcsec, -200*u.arcsec,
                   obstime='2025-09-16', observer="earth", frame=frames.Helioprojective)
    # No error because helioprojective
    _  = attrs.Cutout(hpc, width=900*u.arcsec, height=900*u.arcsec)

    hpr = SkyCoord(500*u.arcsec, -200*u.arcsec,
                   obstime='2025-09-16', observer="earth", frame=frames.HelioprojectiveRadial)
    # Error because not helioprojective
    with pytest.raises(ValueError, match="`bottom_left` must be in the `Helioprojective` frame, but is instead in the `HelioprojectiveRadial` frame"):
        _ = attrs.Cutout(hpr, width=900*u.arcsec, height=900*u.arcsec)


def test_cutout_not_on_disk_when_tracking():
    bottom_left = SkyCoord(500*u.arcsec, -200*u.arcsec,
                           obstime='2025-09-16', observer="earth", frame=frames.Helioprojective)

    # No error because tracking is disabled
    cutout  = attrs.Cutout(bottom_left, width=900*u.arcsec, height=900*u.arcsec, tracking=False)
    assert cutout.value["x"] == 950
    assert cutout.value["y"] == 250

    # Error because tracking is enabled
    with pytest.raises(ValueError, match="Tracking is enabled, but the center of the cutout .* is not on the solar disk"):
        _  = attrs.Cutout(bottom_left, width=900*u.arcsec, height=900*u.arcsec, tracking=True)



"""
This module was built to test the HEK2VSO class.
"""
import numpy as np
import pytest

import astropy.units as u

from sunpy.net import attrs as a
from sunpy.net import hek, hek2vso, vso
from sunpy.time import parse_time

startTime = "2011/08/09 07:23:56"
endTime = "2011/08/09 12:40:29"
eventType = "FL"
instrument = "eit"

hekTime = a.Time(startTime, endTime)
hekEvent = a.hek.EventType(eventType)


@pytest.fixture()
def h2v_client():
    return hek2vso.H2VClient()


@pytest.fixture()
def hek_client():
    return hek.HEKClient()


@pytest.fixture
def vso_client():
    vso.VSOClient()


@pytest.mark.remote_data
def test_translate_results_to_query(hek_client):
    """Make sure that conversion of HEK results to VSO queries is accurate"""
    h = hek_client
    hek_query = h.search(hekTime, hekEvent, a.hek.FL.PeakFlux > 1000)
    vso_query = hek2vso.translate_results_to_query(hek_query)

    # Comparing length of two lists
    assert len(hek_query) == len(vso_query)
    # Comparing types of both queries
    # Not sure this test makes any sense now
    assert isinstance(hek_query, hek.HEKTable)
    assert isinstance(vso_query, list)


@pytest.mark.remote_data
def test_vso_attribute_parse(hek_client):
    """Make sure that Parsing of VSO attributes from HEK queries is accurate"""
    h = hek_client
    hek_query = h.search(hekTime, hekEvent)
    vso_query = hek2vso.vso_attribute_parse(hek_query[0])

    # Checking Time
    assert vso_query[0].start == parse_time(hek_query[0]["event_starttime"])
    assert vso_query[0].end == parse_time(hek_query[0]["event_endtime"])

    # Checking Observatory
    assert vso_query[1].value == hek_query[0]["obs_observatory"]

    # Checking Instrument
    assert vso_query[2].value == hek_query[0]["obs_instrument"]

    # Checking Wavelength
    assert vso_query[3].min == hek_query[0]["obs_meanwavel"]
    assert vso_query[3].max == hek_query[0]["obs_meanwavel"]
    assert vso_query[3].unit == u.Unit("Angstrom")


@pytest.mark.remote_data
def test_members(h2v_client):
    client = h2v_client
    assert isinstance(client.hek_client, hek.HEKClient)
    assert isinstance(client.vso_client, vso.VSOClient)
    assert client.hek_results == ""
    assert client.vso_results == []
    assert client.num_of_records == 0


@pytest.mark.remote_data
def test_translate_and_query(h2v_client, hek_client):
    h = hek_client
    h2v = h2v_client
    q = h.search(a.Time(startTime, endTime),
                 a.hek.EventType(eventType),
                 a.hek.FL.PeakFlux > 1000)
    h2v_q = h2v.translate_and_query(q)

    assert len(q) == len(h2v_q)
    assert isinstance(h2v_q, list)
    assert isinstance(h2v_q[0], vso.VSOQueryResponseTable)


@pytest.mark.remote_data
def test_full_query(h2v_client, hek_client):
    h2v = h2v_client
    h = hek_client
    h2v_q_1 = h2v.full_query(
        (a.Time(startTime, "2011/08/09 07:35"),
         a.hek.EventType(eventType),
         a.hek.FL.PeakFlux > 1000)
    )

    assert h2v.num_of_records > 1
    assert len(h2v.vso_results) > 1
    assert len(h2v.hek_results) > 1

    h2v._quick_clean()
    q = h.search(a.Time(startTime, "2011/08/09 07:35"),
                 a.hek.EventType(eventType),
                 a.hek.FL.PeakFlux > 1000)
    h2v_q_2 = h2v.translate_and_query(q)

    assert len(h2v_q_1) == len(h2v_q_2)
    assert len(h2v.hek_results) == len(q)
    assert h2v.hek_results[0].get_voevent() == q[0].get_voevent()

    for i in range(len(h2v_q_1)):
        assert len(h2v_q_1[i]) == len(h2v_q_2[i])
        if i != 2:
            assert np.nan_to_num(h2v_q_1[i].total_size()) == np.nan_to_num(h2v_q_2[i].total_size())


@pytest.mark.remote_data
def test_quick_clean(h2v_client, hek_client):
    h2v = h2v_client
    h2v.full_query(
        (a.Time(startTime, endTime), a.hek.EventType(eventType), a.hek.FL.PeakFlux > 1000)
    )

    assert h2v.num_of_records != 0
    assert len(h2v.vso_results) != 0

    h2v._quick_clean()

    assert h2v.vso_results == []
    assert h2v.num_of_records == 0



"""
=============================================
Plotting the solar equator and prime meridian
=============================================

This example shows how you can draw the solar equator and prime meridian
(zero Carrington longitude) on Maps as seen by the axes observer.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.data.sample import AIA_171_IMAGE
from sunpy.visualization import drawing

###############################################################################
# Let's import sample AIA data and plot it with the equator and prime meridian.

aia_map = sunpy.map.Map(AIA_171_IMAGE)

fig = plt.figure()
ax = fig.add_subplot(projection=aia_map)
aia_map.plot()
drawing.equator(ax, color='blue')
drawing.prime_meridian(ax, color='red')
plt.show()

###############################################################################
# The draw methods also work with heliogaphic maps. To demonstrate this, let's
# convert the AIA map to a heliographic Carrington map then plot it with the
# equator and prime meridian.

frame_out = SkyCoord(0, 0, unit=u.deg,
                     frame="heliographic_carrington",
                     obstime=aia_map.date,
                     observer=aia_map.observer_coordinate,
                     rsun=aia_map.coordinate_frame.rsun)
header = sunpy.map.make_fitswcs_header((720, 1440),
                                       frame_out,
                                       scale=(360 / 1440,
                                              180 / 720) * u.deg / u.pix,
                                       projection_code="CAR")
outmap = aia_map.reproject_to(header)

fig = plt.figure()
ax = fig.add_subplot(projection=outmap)
outmap.plot(axes=ax)
drawing.equator(ax, color='blue')
drawing.prime_meridian(ax, color='red')
plt.show()



import json
import pathlib
from concurrent.futures import ThreadPoolExecutor

import requests

from astropy.table import Table

from .cdaweb import _CDAS_BASEURL, _CDAS_HEADERS, _DATAVIEW

__all__ = ['get_observatory_groups', 'get_datasets']


def get_observatory_groups():
    """
    Get a list of observatory IDs for each observatory in CDAWeb.

    An observatory group is typically a single mission, which can contain
    multiple observatories, e.g. for the STEREO observatory group there are two
    observatories, STEREO-A and STEREO-B.

    Returns
    -------
    `astropy.table.Table`

    Examples
    --------
    >>> from sunpy.net.cdaweb import get_observatory_groups
    >>>
    >>> groups = get_observatory_groups() # doctest: +REMOTE_DATA
    >>> groups['Group'] # doctest: +REMOTE_DATA
        <Column name='Group' dtype='str55' length=...>
                        ACE
                        AIM
                      AMPTE
        ...
                    Voyager
                       Wind
    >>> groups.loc['STEREO'] # doctest: +REMOTE_DATA
    <Row index=...>
    Group                                  Observatories
    str55                                      str...
    ------ -----------------------------------------------------------------------------
    STEREO 'Ahead', 'Behind', 'STA', 'STB', 'STEREO', 'STEREOA', 'STEREOB', 'sta', 'stb'
    """
    # Get a list of files for a given dataset between start and end times
    url = '/'.join([
        _CDAS_BASEURL,
        'dataviews', _DATAVIEW,
        'observatoryGroups'
    ])
    response = requests.get(url, headers=_CDAS_HEADERS)
    obs_groups = response.json()

    names = [obs['Name'] for obs in obs_groups['ObservatoryGroupDescription']]
    obs_ids = [obs['ObservatoryId'] for obs in obs_groups['ObservatoryGroupDescription']]
    # Join all IDs into a single string
    obs_ids = ["'" + "', '".join(id) + "'" for id in obs_ids]

    t = Table([names, obs_ids], names=['Group', 'Observatories'])
    t.add_index('Group')
    return t


def get_datasets(observatory):
    """
    Get a list of datasets for a given observatory.

    Parameters
    ----------
    observatory : `str`
        Observatory name.

    Returns
    -------
    `astropy.table.Table`

    Examples
    --------
    >>> from sunpy.net.cdaweb import get_datasets
    >>>
    >>> datasets = get_datasets('STEREOB') # doctest: +REMOTE_DATA
    >>> datasets['Id'] # doctest: +REMOTE_DATA
    <Column name='Id' dtype='str17' length=5>
        STB_LB_IMPACT
    STB_L1_IMPACT_HKP
           STB_L1_HET
      STB_L2_SWEA_PAD
     STB_L1_SWEA_SPEC
    >>> datasets.loc['STB_L1_SWEA_SPEC']['Label'] # doctest: +REMOTE_DATA
    np.str_('STEREO Behind IMPACT/SWEA Spectra - ...')
    >>> datasets.loc['STB_L1_SWEA_SPEC'][['Start', 'End']] # doctest: +REMOTE_DATA
    <Row index=4>
             Start                     End
             str24                    str24
    ------------------------ ------------------------
    2012-12-01T00:00:03.000Z 2013-12-31T23:59:41.000Z
    """
    # Get a list of files for a given dataset between start and end times
    url = '/'.join([
        _CDAS_BASEURL,
        'dataviews', _DATAVIEW,
        'datasets'
    ])
    url = f'{url}?observatory={observatory}'
    response = requests.get(url, headers=_CDAS_HEADERS)
    datasets = response.json()['DatasetDescription']

    ids = [dataset['Id'] for dataset in datasets]
    instruments = [', '.join(dataset['Instrument']) for dataset in datasets]
    labels = [dataset['Label'] for dataset in datasets]
    stimes = [dataset['TimeInterval']['Start'] for dataset in datasets]
    etimes = [dataset['TimeInterval']['End'] for dataset in datasets]

    t = Table([ids, instruments, labels, stimes, etimes],
              names=['Id', 'Instruments', 'Label', 'Start', 'End'])
    t.add_index('Id')
    return t


def _update_cdaweb_dataset_data():
    all_obs = get_observatory_groups()
    url = '/'.join([
        _CDAS_BASEURL,
        'dataviews', _DATAVIEW,
        'datasets'
    ])
    # Mapping from dataset ID to description
    all_datasets = {}
    # Number of parallel threads we spawn
    N = 3

    def _fetch_cdaweb_dataset(group, url=url):
        print(f'🛰 Getting datasets for {group}')
        u = url + f'?observatoryGroup={group}'
        res = requests.get(u, headers=_CDAS_HEADERS)
        datasets = res.json()['DatasetDescription']
        dataset_ids = {ds['Id']: ds['Label'] for ds in datasets}
        all_datasets.update(dataset_ids)

    with ThreadPoolExecutor(max_workers=N) as executor:
        # Submit each URL to the thread pool
        futures = [executor.submit(_fetch_cdaweb_dataset, group) for group in all_obs['Group']]
        # Wait for all tasks to complete
        for future in futures:
            future.result()

    attr_file = pathlib.Path(__file__).parent / 'data' / 'attrs.json'
    with open(attr_file, 'w') as attrs_file:
        json.dump(dict(sorted(all_datasets.items())), attrs_file, indent=2)



import os
from os.path import basename
from collections import OrderedDict

import numpy as np
from pandas import DataFrame, to_datetime
from pandas.io.parsers import read_csv

import astropy.units as u
from astropy.time import TimeDelta

import sunpy.io
import sunpy.io._file_tools
from sunpy.time import parse_time
from sunpy.timeseries.timeseriesbase import GenericTimeSeries
from sunpy.util.metadata import MetaDict
from sunpy.visualization import peek_show

__all__ = ['EVESpWxTimeSeries', 'ESPTimeSeries']


class ESPTimeSeries(GenericTimeSeries):
    """
    SDO EVE/ESP Level 1 data.

    The Extreme ultraviolet Spectro-Photometer (ESP) is an irradiance instrument
    which is part of the Extreme ultraviolet Variability Experiment (EVE) on board
    SDO. ESP provides high time cadence (0.25s) EUV irradiance measurements in five
    channels, one soft X-ray and 4 EUV. The first four orders of the diffraction grating
    gives measurements centered on 18nm, 26nm, 30nm and 36nm. The zeroth order (obtained
    by 4 photodiodes) provides the soft X-ray measurements from 0.1-7nm.

    The ESP level 1 fits files are fully calibrated. The TimeSeries object created from
    an ESP fits file will contain 4 columns namely:

        * 'QD' - sum of 4 quad diodes, this is the soft X-ray measurements 0.1-7nm
        * 'CH_18' - EUV irradiance 18nm
        * 'CH_26' - EUV irradiance 26nm
        * 'CH_30' - EUV irradiance 30nm
        * 'CH_36' - EUV irradiance 36nm

    .. note::

        The 36 nm channel demonstrates a significant noise and it is not recommended to be
        used for short-time observations of solar irradiance.

    References
    ----------
    * `SDO Mission Homepage <https://sdo.gsfc.nasa.gov/>`__
    * `EVE Homepage <https://lasp.colorado.edu/eve/>`__
    * `README ESP data <https://lasp.colorado.edu/eve/data_access/evewebdata/products/level1/esp/EVE_ESP_L1_V6_README.pdf>`__
    * `ESP lvl1 data <https://lasp.colorado.edu/eve/data_access/evewebdata/misc/eve_calendars/calendar_level1_2018.html>`__
    * ESP instrument paper: :cite:t:`didkovsky_euv_2012`
    """

    _source = 'esp'
    _url = "https://lasp.colorado.edu/eve/"

    def plot(self, axes=None, columns=None, **kwargs):
        """
        Plots the EVE ESP Level 1 timeseries data.

        Parameters
        ----------
        axes : numpy.ndarray of `matplotlib.axes.Axes`, optional
            The axes on which to plot the TimeSeries.
        columns : list[str], optional
            If provided, only plot the specified columns.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to `pandas.DataFrame.plot`.

        Returns
        -------
        array of `~matplotlib.axes.Axes`
            The plot axes.
        """
        axes, columns = self._setup_axes_columns(axes, columns, subplots=True)
        column_names = {"QD": "Flux \n 0.1-7nm", "CH_18": "Flux \n 18nm",
                        "CH_26": "Flux \n 26nm", "CH_30": "Flux \n 30nm", "CH_36": "Flux \n 36nm"}

        for i, name in enumerate(self.to_dataframe()[columns]):
            axes[i].plot(self._data[name],
                         label=name)
            axes[i].set_ylabel(column_names[name])
            axes[i].legend(loc="upper right")
        axes[-1].set_xlim(self._data.index[0], self._data.index[-1])
        self._setup_x_axis(axes)
        return axes

    @peek_show
    def peek(self, *, title="EVE/ESP Level 1", columns=None, **kwargs):
        """
        Displays the EVE ESP Level 1 timeseries data by calling
        `~sunpy.timeseries.sources.eve.ESPTimeSeries.plot`.

        Parameters
        ----------
        title : `str`, optional
            The title of the plot. Defaults to "EVE/ESP Level 1".
        columns : list[str], optional
            If provided, only plot the specified columns.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to `pandas.DataFrame.plot`.
        """
        axes = self.plot(columns=columns, **kwargs)
        axes[0].set_title(title)
        fig = axes[0].get_figure()
        fig.tight_layout()
        fig.subplots_adjust(hspace=0.05)
        return fig

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses a EVE ESP level 1 data.
        """
        hdus = sunpy.io._file_tools.read_file(filepath)
        return cls._parse_hdus(hdus)

    @classmethod
    def _parse_hdus(cls, hdulist):
        header = MetaDict(OrderedDict(hdulist[0].header))
        # Adding telescope to MetaData
        header.update({'TELESCOP': hdulist[1].header['TELESCOP'].split()[0]})

        start_time = parse_time(hdulist[1].header['T_OBS'])
        times = start_time + TimeDelta(hdulist[1].data['SOD']*u.second)

        colnames = ['QD', 'CH_18', 'CH_26', 'CH_30', 'CH_36']

        all_data = [hdulist[1].data[x] for x in colnames]
        data = DataFrame(np.array(all_data).T, index=times.isot.astype(
            'datetime64'), columns=colnames)
        data.sort_index(inplace=True)

        units = OrderedDict([('QD', u.W/u.m**2),
                             ('CH_18', u.W/u.m**2),
                             ('CH_26', u.W/u.m**2),
                             ('CH_30', u.W/u.m**2),
                             ('CH_36', u.W/u.m**2)])

        return data, header, units

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if header corresponds to an EVE image.
        """
        if kwargs.get('source', ''):
            return kwargs.get('source', '').lower().startswith(cls._source)
        if 'meta' in kwargs.keys():
            return kwargs['meta'].get('TELESCOP', '').endswith('SDO/EVE')


class EVESpWxTimeSeries(GenericTimeSeries):
    """
    SDO EVE LightCurve for level 0CS data.

    The Extreme Ultraviolet Variability Experiment (EVE) is an instrument on board the Solar Dynamics Observatory (SDO).
    The EVE instrument is designed to measure the solar extreme ultraviolet (EUV) irradiance.
    The EUV radiation includes the 0.1-105 nm range, which provides the majority
    of the energy for heating Earth's thermosphere and creating Earth's ionosphere (charged plasma).

    EVE includes several irradiance instruments:

    * The Multiple EUV Grating Spectrographs (MEGS)-A is a grazing- incidence spectrograph
      that measures the solar EUV irradiance in the 5 to 37 nm range with 0.1-nm resolution,
    * The MEGS-B is a normal-incidence, dual-pass spectrograph that measures the solar EUV
      irradiance in the 35 to 105 nm range with 0.1-nm resolution.

    Level 0CS data is primarily used for space weather.
    It is provided near real-time and is crudely calibrated 1-minute averaged broadband irradiances from ESP and MEGS-P broadband.
    For other levels of EVE data, use `~sunpy.net.Fido`, with ``sunpy.net.attrs.Instrument('eve')``.

    Data is available starting on 2010/03/01.

    Examples
    --------
    >>> import sunpy.timeseries
    >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
    >>> eve = sunpy.timeseries.TimeSeries(sunpy.data.sample.EVE_TIMESERIES, source='EVE')  # doctest: +REMOTE_DATA
    >>> eve = sunpy.timeseries.TimeSeries("https://lasp.colorado.edu/eve/data_access/evewebdata/quicklook/L0CS/LATEST_EVE_L0CS_DIODES_1m.txt", source='EVE')  # doctest: +SKIP
    >>> eve.peek(subplots=True)  # doctest: +SKIP

    References
    ----------
    * `SDO Mission Homepage <https://sdo.gsfc.nasa.gov/>`__
    * `EVE Homepage <https://lasp.colorado.edu/eve/>`__
    * `Level 0CS Definition <https://lasp.colorado.edu/eve/data/>`__
    * `EVE Data Access <https://lasp.colorado.edu/eve/data_access/index.html>`__
    * Instrument Paper :cite:t:`woods_extreme_2012`.
    """
    # Class attribute used to specify the source class of the TimeSeries.
    _source = 'eve'
    _url = "https://lasp.colorado.edu/eve/"

    @peek_show
    def peek(self, *, columns=None, **kwargs):
        """
        Plots the time series in a new figure.

        .. plot::

            import sunpy.timeseries
            from sunpy.data.sample import EVE_TIMESERIES
            eve = sunpy.timeseries.TimeSeries(EVE_TIMESERIES, source='eve')
            eve.peek(subplots=True, figsize=(22,11))

        Parameters
        ----------
        columns : list[str], optional
            If provided, only plot the specified columns.
        **kwargs : `dict`
            Additional plot keyword arguments that are handed to
            :meth:`pandas.DataFrame.plot`.
        """
        # Check we have a timeseries valid for plotting
        self._validate_data_for_plotting()

        # Choose title if none was specified
        if "title" not in kwargs and columns is None:
            if len(self.to_dataframe().columns) > 1:
                kwargs['title'] = 'EVE (1 minute data)'
            else:
                if self._filename is not None:
                    base = self._filename.replace('_', ' ')
                    kwargs['title'] = os.path.splitext(base)[0]
                else:
                    kwargs['title'] = 'EVE Averages'

        if columns is None:
            axes = self.to_dataframe().plot(sharex=True, **kwargs)
        else:
            data = self.to_dataframe()[columns]
            if "title" not in kwargs and len(columns) == 1:
                kwargs['title'] = 'EVE ' + columns[0].replace('_', ' ')
            else:
                kwargs['title'] = 'EVE Averages'
            axes = data.plot(sharex=True, **kwargs)

        if "subplots" in kwargs:
            fig = axes[0].get_figure()
        else:
            fig = axes.get_figure()
        return fig

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses an EVE CSV file.
        """
        cls._filename = basename(filepath)
        with open(filepath, encoding='ascii') as fp:
            # Determine type of EVE CSV file and parse
            line1 = fp.readline()

        if line1.startswith("Date"):
            raise NotImplementedError("Reading SDO/EVE level 0CS average files is not implemented.")
        elif line1.startswith(";"):
            return cls._parse_level_0cs(filepath)

    @staticmethod
    def _parse_level_0cs(filepath):
        """
        Parses and EVE Level 0CS file.
        """
        is_missing_data = False  # boolean to check for missing data
        missing_data_val = np.nan
        header = []
        fields = []
        with open(filepath, encoding='ascii') as fp:
            line = fp.readline()
            # Read header at top of file
            while line.startswith(";"):
                header.append(line)
                if '; Missing data:' in line:
                    is_missing_data = True
                    missing_data_val = line.split(':')[1].strip()

                line = fp.readline()

        meta = MetaDict()
        for hline in header:
            if hline == '; Format:\n' or hline == '; Column descriptions:\n':
                continue
            elif ('Created' in hline) or ('Source' in hline):
                meta[hline.split(':',
                                 1)[0].replace(';',
                                               ' ').strip()] = hline.split(':', 1)[1].strip()
            elif ':' in hline:
                meta[hline.split(':')[0].replace(';', ' ').strip()] = hline.split(':')[1].strip()

        fieldnames_start = False
        for hline in header:
            if hline.startswith("; Format:"):
                fieldnames_start = False
            if fieldnames_start:
                fields.append(hline.split(":")[0].replace(';', ' ').strip())
            if hline.startswith("; Column descriptions:"):
                fieldnames_start = True

        # Next line is YYYY DOY MM DD
        date_parts = line.split(" ")
        year = int(date_parts[0])
        month = int(date_parts[2])
        day = int(date_parts[3])

        data = read_csv(filepath, sep=r'\s+', names=fields, comment=';', dtype={'HHMM': int})
        # First line is YYYY DOY MM DD
        data = data.iloc[1:, :]
        data['Hour'] = data['HHMM'] // 100
        data['Minute'] = data['HHMM'] % 100
        data = data.drop(['HHMM'], axis=1)

        data['Year'] = year
        data['Month'] = month
        data['Day'] = day

        datecols = ['Year', 'Month', 'Day', 'Hour', 'Minute']
        data['Time'] = to_datetime(data[datecols])
        data = data.set_index('Time')
        data = data.drop(datecols, axis=1)

        if is_missing_data:  # If missing data specified in header
            data[data == float(missing_data_val)] = np.nan

        # Add the units data
        units = OrderedDict([('XRS-B proxy', u.W/u.m**2),
                             ('XRS-A proxy', u.W/u.m**2),
                             ('SEM proxy', u.W/u.m**2),
                             ('0.1-7ESPquad', u.W/u.m**2),
                             ('17.1ESP', u.W/u.m**2),
                             ('25.7ESP', u.W/u.m**2),
                             ('30.4ESP', u.W/u.m**2),
                             ('36.6ESP', u.W/u.m**2),
                             ('darkESP', u.ct),
                             ('121.6MEGS-P', u.W/u.m**2),
                             ('darkMEGS-P', u.ct),
                             ('q0ESP', u.dimensionless_unscaled),
                             ('q1ESP', u.dimensionless_unscaled),
                             ('q2ESP', u.dimensionless_unscaled),
                             ('q3ESP', u.dimensionless_unscaled),
                             ('CMLat', u.deg),
                             ('CMLon', u.deg)])
        # Todo: check units used.
        return data, meta, units

    @classmethod
    def is_datasource_for(cls, **kwargs):
        """
        Determines if header corresponds to an EVE image.
        """
        if kwargs.get('source', ''):
            return kwargs.get('source', '').lower().startswith(cls._source)



"""
This module provides plotting support in iPython.
"""
from functools import wraps

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.transforms import Transform

import astropy.units as u
from astropy.visualization.wcsaxes import CoordinateHelper
from astropy.wcs.utils import pixel_to_pixel

__all__ = ['peek_show', "axis_labels_from_ctype", "show_hpr_impact_angle"]


def peek_show(func):
    """
    A decorator to place on ``peek()`` methods to show the figure.

    The ``peek()`` method should return the figure then this method will
    attempt to show it in the correct way. This decorator will not return the
    figure to the user.
    """
    @wraps(func)
    def show_figure(*args, **kwargs):
        _ = func(*args, **kwargs)
        plt.show()

    return show_figure


def axis_labels_from_ctype(ctype, unit):
    """
    Returns axis labels for the given coordinate type and unit.

    Parameters
    ----------
    ctype : `str`
        Coordinate type.
    unit : `str`, `None`
        Required unit. If `None` no unit is added to the label.

    Returns
    -------
    `str`
        "Axis Label [Unit]"
    """
    ctype_short = ctype[:4]

    labels = {'HGLN': 'Heliographic Longitude',
              'CRLN': 'Carrington Longitude',
              'HPLN': 'Helioprojective Longitude (Solar-X)',
              'HRLN': 'Helioprojective Position Angle',
              'SOLX': 'Heliocentric X',

              'HGLT': 'Latitude',
              'CRLT': 'Latitude',
              'HPLT': 'Helioprojective Latitude (Solar-Y)',
              'HRLT': 'Helioprojective Declination',
              'SOLY': 'Heliocentric Y'}

    label = labels.get(ctype_short, f"{ctype}")
    if unit is not None:
        label += f' [{unit}]'

    return label


def show_hpr_impact_angle(declination_axis):
    """
    Modify a declination axis for Helioprojective Radial plots to show impact angle
    instead.

    In order to accommodate the FITS-WCS machinery, the declination is used instead
    of the impact angle (which is the declination plus 90 degrees). This function
    changes the tick-label formatter for an axis so that it adds 90 degrees before
    rendering the labels.

    Parameters
    ----------
    coordinate_axis : `astropy.visualization.wcsaxes.CoordinateHelper`
        The coordinate axis for Helioprojective Radial declination (``delta``)

    See Also
    --------
    sunpy.coordinates.HelioprojectiveRadial

    Examples
    --------
    .. minigallery:: sunpy.visualization.show_hpr_impact_angle
    """
    if not isinstance(declination_axis, CoordinateHelper):
        raise TypeError("The input should be one of the WCSAxes coordinate axes.")

    old_format_method = declination_axis.formatter

    def new_format_method(values, spacing, format="auto"):
        return old_format_method(values + 90*u.deg, spacing, format=format)
    # TODO: Astropy >= 7.0 use declination_axis.set_major_formatter(new_format_method)
    declination_axis._formatter_locator.formatter = new_format_method
    declination_axis.set_axislabel('Helioprojective Impact Angle')


# TODO: Consider making this transform class public
class _PrecomputedPixelCornersTransform(Transform):
    """
    A matplotlib transform that precomputes the pixel->pixel transformation from a
    given data WCS to the WCS of the given WCSAxes, and stores it as a lookup table
    to avoid redundant coordinate transformations. This transformation is computed
    only for the pixel corners of the data WCS (e.g., (-0.5, -0.5) for the bottom-
    left corner), so the transform may not return desired values when called with
    pixel coordinates other than at the corners of pixels.
    """
    input_dims = 2
    output_dims = 2

    def __init__(self, axes, data_wcs):
        super().__init__()
        self.data_y, self.data_x = np.indices(np.array(data_wcs.array_shape) + 1) - 0.5
        self.axes_x, self.axes_y = pixel_to_pixel(data_wcs, axes.wcs, self.data_x, self.data_y)

    def transform_non_affine(self, values):
        ix, iy = np.ceil(values).astype(int).T
        return np.stack([self.axes_x[iy, ix], self.axes_y[iy, ix]], axis=1)
